from dataclasses import dataclass
import gzip, random
from typing import Any, List, Sequence, Union
from ..locale import Lang
from ..traffic import EV,Trip
from feasytools.pdf import *


@dataclass
class VehicleType:
    id:int
    bcap_kWh:float
    range_km:float
    efc_rate_kW:float
    esc_rate_kW:float
    max_V2G_kW:float
    
def random_diff(seq:Sequence[Any], exclude:Any):
    """
    Choose a random element from `seq` that is not equal to `exclude`
    """
    ret = exclude
    if len(seq) == 1 and seq[0] == exclude:
        raise RuntimeError(Lang.ERROR_RANDOM_CANNOT_EXCLUDE)
    while ret == exclude:
        ret = random.choice(seq)
    return ret

class _TripInner:
    def __init__(self, trip_id:str, depart_time:Union[str,int], from_node:str, from_type:str,
            to_node:str, to_type:str,):
        '''
        Initialize a trip object.
            trip_id: Unique trip ID
            depart_time: Departure time in seconds since midnight of the day.
                Can be a string or an integer. If it is a string, it should be convertible to an integer.
            from_node: Origin node ID
            from_type: The type of place where the trip starts.
            to_node: Destination node ID
            to_type: The type of place where the trip will end.
        '''
        self.id = trip_id
        self.DPTT = int(depart_time) # departure time in seconds since midnight
        self.frN = from_node
        self.toN = to_node
        self.frT = from_type
        self.toT = to_type
    
    def toXML(self, daynum:int) -> str:
        return (f'\n<trip id="{self.id}" depart="{self.DPTT + 86400 * daynum}" ' + 
            f'fromType="{self.frT}" toType="{self.toT}" ' + 
            f'fromNode="{self.frN}" toNode="{self.toN}" />')
    
    def toTrip(self, daynum:int) -> Trip:
        return Trip(self.id, self.DPTT + 86400 * daynum, self.frN, self.toN)

PDFuncLike = Union[None, float, PDFunc]
def _impl_PDFuncLike(x:PDFuncLike, default:PDFunc) -> float:
    if x is None:
        return default.sample()
    elif isinstance(x, float):
        return x
    elif isinstance(x, PDFunc):
        return x.sample()
    raise TypeError("x must be None, float or PDFunc")
    
class _EVInner:
    """
    EV class used to generate trips
    """

    def __init__(self, veh_id: str, vT:VehicleType, soc:float, v2g_prop:float = 1.0+1e-4, 
        omega:PDFuncLike = None, krel:PDFuncLike = None,
        ksc:PDFuncLike = None, kfc:PDFuncLike = None, 
        kv2g:PDFuncLike = None, cache_route:bool = False,
    ):
        '''
        Initialize EV object
            veh_id: Vehicle ID
            vT: Vehicle type
            soc: State of charge
            v2g_prop: Vehicle's probability of being able to V2G. Value >= 1.0 means always V2G.
            omega: PDF for omega. None for random uniform between 5 and 10.
                omega indicates the user's sensitivity to the cost of charging. Bigger omega means less sensitive.
            krel: PDF for krel. None for random uniform between 1 and 1.2.
                krel indicates the user's estimation of the distance. Bigger krel means the user underestimates the distance.
            ksc: PDF for ksc. None for random uniform between 0.4 and 0.6.
                ksc indicates the SoC threshold for slow charging.
            kfc: PDF for kfc. None for random uniform between 0.2 and 0.25.
                kfc indicates the SoC threshold for fast charging halfway.
            kv2g: PDF for kv2g. None for random uniform between 0.65 and 0.75.
                kv2g indicates the SoC threshold of the battery that can be used for V2G.'
            cache_route: Wheter remember route for further use.
        '''
        self.vehicle_id = veh_id
        self.bcap = vT.bcap_kWh
        self.soc = soc
        self.consump_Whpm = vT.bcap_kWh / vT.range_km  # kWh/km = Wh/m
        self.efc_rate_kW = vT.efc_rate_kW
        self.esc_rate_kW = vT.esc_rate_kW
        self.max_v2g_rate_kW = vT.max_V2G_kW
        self.omega = _impl_PDFuncLike(omega, PDUniform(5.0, 10.0))
        self.krel = _impl_PDFuncLike(krel, PDUniform(1.0, 1.2))
        self.ksc = _impl_PDFuncLike(ksc, PDUniform(0.4, 0.6))
        self.kfc = _impl_PDFuncLike(kfc, PDUniform(0.2, 0.25))
        self.cache_route = cache_route
        if v2g_prop >= 1.0 or random.random() < v2g_prop:
            self.kv2g = _impl_PDFuncLike(kv2g, PDUniform(0.65, 0.75))
        else:
            self.kv2g = 1 + 1e-4
        self.trips:List[_TripInner] = []
        self.daynum:List[int] = []
    
    def _add_trip(self, daynum: int, trip_dict: _TripInner):
        self.daynum.append(daynum)
        self.trips.append(trip_dict)
    
    def addTrip(self, trip_id:str, depart_time:int, from_node:str, from_Type:str,
                to_node:str, to_Type:str, daynum:int = -1):
        '''
        Add a trip to the EV.
            trip_id: Unique trip ID
            depart_time: Departure time in seconds since midnight of the day.
                When it is less than 86400 and `daynum` >= 0, it is considered as in the day specified by `daynum`.
                Otherwise, it is considered as the exact time in seconds since simulation start.
            from_TAZ: Origin TAZ ID
            from_EDGE: Origin edge ID
            to_TAZ: Destination TAZ ID
            to_EDGE: Destination edge ID
            route: List of edge IDs representing the route, should have at least 2 elements.
                When the route has only 2 elements, it is considered only the start and end edges.
            fixed_route: Whether the route is fixed or not.
                True means the route is fixed.
                False means the route is not fixed and must be recalculated every time the trip is used.
                None means it is not specified. Whether the route is fixed or not will be determined by the simulation.
            daynum: Day number, starting from 0. -1 means the trip is not in a specific day and `depart_time` is the exact time since simulation start.
        Raises:
            AssertionError: If `daynum` is specified and `depart_time` is not less than 86400.
        '''
        if daynum < 0:
            daynum = depart_time // 86400
            depart_time = depart_time % 86400
        else:
            assert 0 <= depart_time and depart_time < 86400, "When daynum is specified, depart_time must be less than 86400."
        self.daynum.append(daynum)
        depart_time = int(depart_time) if isinstance(depart_time, str) else depart_time
        self.trips.append(_TripInner(trip_id, depart_time, from_node, from_Type, to_node, to_Type))

    def toXML(self) -> str:
        ret = (
            f'<vehicle id="{self.vehicle_id}" soc="{self.soc:.4f}" bcap="{self.bcap:.4f}" c="{self.consump_Whpm:.8f}"'
            + f' rf="{self.efc_rate_kW:.4f}" rs="{self.esc_rate_kW:.4f}" rv="{self.max_v2g_rate_kW:.4f}" omega="{self.omega:.6f}"'
            + f'\n  kf="{self.kfc:.4f}" ks="{self.ksc:.4f}" kv="{self.kv2g:.4f}" kr="{self.krel:.4f}"'
            + f' eta_c="0.9" eta_d="0.9" rmod="Linear" cache_route="{self.cache_route}">'
        )
        for d, tr in zip(self.daynum, self.trips):
            ret += tr.toXML(d)
        ret += "\n</vehicle>\n"
        return ret

    def toEV(self) -> EV:
        trips = [m.toTrip(daynum) for m, daynum in zip(self.trips, self.daynum)]
        return EV(
            self.vehicle_id,
            trips,
            0.9,
            0.9,
            self.bcap,
            self.soc,
            self.consump_Whpm,
            self.efc_rate_kW,
            self.esc_rate_kW,
            self.max_v2g_rate_kW,
            self.omega,
            self.krel,
            self.kfc,
            self.ksc,
            self.kv2g,
            "Linear",
            cache_route=self.cache_route,
        )


class _xmlSaver:
    """Class used to save XML files"""

    def __init__(self, path: str):
        if path.endswith(".gz"):
            self.a = gzip.open(path, "wt")
        else:
            self.a = open(path, "w")
        self.a.write("<root>\n")

    def write(self, e: _EVInner):
        self.a.write(e.toXML())

    def close(self):
        self.a.write("</root>")
        self.a.close()


from fpowerkit import Generator
from feasytools import ComFunc, LangLib
from .pdn import PluginPDN
from .base import *

V2GRes = List[float]

_locale = LangLib(["zh_CN","en"])
_locale.SetLangLib("zh_CN",
    DESCRIPTION = "V2G调度系统",
    ERROR_NO_PDN = "V2G调度依赖于PDN插件",
    ERROR_SMART_CHARGE = "启用有序充电时V2G调度插件不可用",
)
_locale.SetLangLib("en",
    DESCRIPTION = "V2G scheduling system",
    ERROR_NO_PDN = "V2G scheduling depends on PDN plugin",
    ERROR_SMART_CHARGE = "V2G scheduling plugin is not available when smart charging is enabled",
)

class PluginV2G(PluginBase[V2GRes]):
    @property
    def Description(self)->str:
        return _locale["DESCRIPTION"]
    
    def _save_state(self) -> object:
        '''Save the plugin state'''
        return None
    
    def _load_state(self, state:object) -> None:
        '''Load the plugin state'''

    @staticmethod
    def ElemShouldHave() -> ConfigDict:
        '''Get the plugin configuration item list'''
        return ConfigDict()
    
    def Init(self,elem:ET.Element,inst:TrafficInst,work_dir:Path,res_dir:Path,plg_deps:'List[PluginBase]')->V2GRes:
        self.__inst = inst
        self.__step_len = inst.step_len
        self.SetPreStep(self._work)
        self.SetPostStep(self._work_post)

        assert len(plg_deps) == 1 and isinstance(plg_deps[0], IGridPlugin), _locale["ERROR_NO_PDN"]
        self.__pdn = plg_deps[0]
        if isinstance(self.__pdn, PluginPDN) and self.__pdn.isSmartChargeEnabled():
            raise RuntimeError(_locale["ERROR_SMART_CHARGE"])
        self.__inst = inst
        self._cap:List[float] = [0.] * len(inst.SCSList)

        for i,pk in enumerate(inst.SCSList):
            self.__pdn.Grid.AddGen(Generator("V2G_"+pk.name,pk._bus,0.,pk.psell*(self.__pdn.Grid.Sb*1000),0.,
                0.,ComFunc(self.__get_cap(i)),0.,0.,))
        if isinstance(self.__pdn, PluginPDN):
            self.__pdn.Solver.est.UpdateGrid(self.__pdn.Grid)
        return []
    
    def __get_cap(self,i:int):
        def func(t: int)->float:
            if not self.IsOnline(t): return 0.
            return self._cap[i]
        return func
    
    def _work_post(self, _t:int, /, sta:PluginStatus)->Tuple[bool,List[float]]:
        if sta == PluginStatus.EXECUTE or (sta == PluginStatus.OFFLINE and self.IsOnline(_t + self.__step_len)):
            self._cap = [x*3.6/self.__pdn.Grid.Sb for x in self.__inst.SCSList.get_V2G_cap(_t)]
            ret = True, self._cap
        elif sta == PluginStatus.OFFLINE:
            ret = True, []
        elif sta == PluginStatus.HOLD:
            ret = True, self.LastPreStepResult
        return ret
    
    def _work(self,_t:int,/,sta:PluginStatus)->Tuple[bool,List[float]]:
        '''
        Get the V2G demand power of all bus with slow charging stations at time _t, unit kWh/s, 3.6MW=3600kW=1kWh/s
        '''
        if sta == PluginStatus.EXECUTE:
            if self.__pdn.LastPreStepSucceed:
                f = lambda x: (0.0 if x is None else x)*self.__pdn.Grid.Sb/3.6
                ret1 = [f(self.__pdn.Grid.Gen("V2G_"+pk.name).P) for pk in self.__inst.SCSList]
                if sum(ret1)>1e-8: ret = True, ret1
                else: ret = False, []
            else:
                ret = False, []
            self.__inst.SCSList.set_V2G_demand(ret[1])
        elif sta == PluginStatus.OFFLINE:
            self.__inst.SCSList.set_V2G_demand([])
            ret = True, []
        elif sta == PluginStatus.HOLD:
            ret = True, self.LastPreStepResult
        return ret


from abc import abstractmethod
from typing import Any, Iterable, Optional, List, Dict
from pathlib import Path
from ..plugins import *
from ..traffic import TrafficInst

def cross_list(a:Iterable[str],b:Iterable[str])->List[str]:
    '''Generate cross table header'''
    return [f"{i}#{j}" for j in b for i in a]

_DIGITS = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
def to_base62(num:int):
    if num == 0: return '0'
    result = ''
    while num:
        num, remainder = divmod(num, 62)
        result += _DIGITS[remainder]
    return result

class StaBase:
    '''Base class for statistics recorder'''
    @abstractmethod
    def __init__(self, name:str, path:str, items:List[str], tinst:TrafficInst, plugins:Dict[str,PluginBase], 
            precision:Optional[Dict[str, int]]=None, compress:bool=True):
        self._name=name
        self._inst=tinst
        self._plug=plugins
        self._cols=items
        self._vals=[None] * len(items)
        self._writer=open(str(Path(path)/(self._name+".csv")),"w",buffering=1024*1024)
        self._mp = None
        self._pre = precision if precision is not None else {}
        if compress:
            self._mp = {i:to_base62(j) for j,i in enumerate(items)}
            self._writer.write("C\n")
            self._writer.write(','.join(items)+"\n")        
        self._writer.write("Time,Item,Value\n")
        self._lastT = -1
        self.__cnt = 0
    
    @staticmethod
    @abstractmethod
    def GetLocalizedName() -> str:
        '''Get Localized Name'''
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return []
    
    @property
    def Writer(self):
        return self._writer
    
    def GetData(self, inst:TrafficInst, plugins:Dict[str,PluginBase]) -> Iterable[Any]: 
        '''Get Data'''
        raise NotImplementedError
    
    def LogOnce(self):
        data = self.GetData(self._inst, self._plug)
        n = len(self._vals)
        for i, v in enumerate(data):
            if i >= n: raise ValueError(f"{self._name}: Data length ({i+1}) > Column count ({n}).")
            if self._vals[i] is None or abs(v-self._vals[i])>1e-6:
                v = round(v, self._pre.get(self._cols[i], 6))
                if self._mp is not None:
                    col = self._mp[self._cols[i]]
                else:
                    col = self._cols[i]
                if self._lastT != self._inst.current_time:
                    self._lastT = self._inst.current_time
                    self.__cnt += self._writer.write(f"{self._lastT},{col},{v}\n")
                else:
                    self.__cnt += self._writer.write(f",{col},{v}\n")
                self._vals[i] = v
        if n > 0 and i != n - 1: f"{self._name}: Data length ({i+1}) != Column count ({n})."
        if self.__cnt >= 1024*1024:
            self._writer.flush()
            self.__cnt = 0

    def close(self):
        self._writer.close()
    
    def __exit__(self):
        self.close()


from v2simux_gui.common import *

import datetime
from ..mainbox.controls import SelectItemDialog

_L = LangLib.Load(__file__)

def get_clog_mtime(folder:Path):
    clog_path = folder / "cproc.clog"
    if clog_path.is_file():
        return clog_path.stat().st_mtime
    return None


def format_time(ts):
    if ts is None:
        return _L("NOT_FOUND")
    return datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H:%M:%S")


class SelectResultsDialog(SelectItemDialog):
    def __init__(self, items:List[Path]):
        new_items = []
        self.__folders:Dict[str, Path] = {}
        for item in items:
            mtime = get_clog_mtime(item)
            new_items.append([item.name, format_time(mtime)])
            self.__folders[item.name] = item.absolute()
        super().__init__(new_items, title=_L("SELECT_RESULTS"), 
                         columns=[("folder",_L("FOLDER")), ("mtime",_L("MODIFIED_TIME"))])
        self.tree.column("folder", width=300)
        self.tree.column("mtime", width=180)
    
    @property
    def folder(self) -> Union[Path, None]:
        if self.selected_item is None:
            return None
        return self.__folders[self.selected_item[0]]
__all__ = ["SelectResultsDialog", "get_clog_mtime", "format_time"]


from v2simux_gui.com_no_vx import *
from v2simux_gui.langhelper import *

import os
from PIL import Image, ImageTk


_ = LangLib.Load(__file__)


class CmpBox(Tk):
    def __init__(self):
        super().__init__()

        self.title(_("TITLE"))
        self.geometry("1024x768")
        self.bind("<Configure>", self.on_resize)

        self.folder1 = None
        self.folder2 = None
        self.original_image1 = None
        self.original_image2 = None
        self.folder_buf = "./results"

        self.create_widgets()

    def create_widgets(self):
        self.menu = Menu(self)
        self.config(menu=self.menu)
        self.filemenu = Menu(self.menu, tearoff=0)
        self.filemenu.add_command(label=_("OPEN_FOLDER1"), command=self.open_folder1)
        self.filemenu.add_command(label=_("OPEN_FOLDER2"), command=self.open_folder2)
        self.filemenu.add_separator()
        self.filemenu.add_command(label=_("EXIT"), command=self.destroy)
        self.menu.add_cascade(label=_("FILE"), menu=self.filemenu)
        add_lang_menu(self.menu)

        self.sidebar = Frame(self)
        self.sidebar.pack(side=LEFT, fill=Y)

        self.folder1_button = Button(self.sidebar, text=_("OPEN_FOLDER1"), command=self.open_folder1)
        self.folder1_button.pack(pady=10)

        self.folder2_button = Button(self.sidebar, text=_("OPEN_FOLDER2"), command=self.open_folder2)
        self.folder2_button.pack(pady=10)

        self.folder1_label = Label(self.sidebar, text=_("LB_FOLDER1").format(_("TO_BE_SELECTED")))
        self.folder1_label.pack(pady=10)

        self.folder2_label = Label(self.sidebar, text=_("LB_FOLDER2").format(_("TO_BE_SELECTED")))
        self.folder2_label.pack(pady=10)

        self.file_listbox = Listbox(self.sidebar)
        self.file_listbox.pack(fill=BOTH, expand=True, pady=10)
        self.file_listbox.bind("<<ListboxSelect>>", self.on_file_select)

        self.image_frame = Frame(self)
        self.image_frame.pack(side=RIGHT, fill=X, expand=True)

        self.image1_label = Label(self.image_frame)
        self.image1_label.pack(side=TOP, padx=10, pady=10)

        self.image2_label = Label(self.image_frame)
        self.image2_label.pack(side=TOP, padx=10, pady=10)

    def open_folder1(self):
        new_folder = filedialog.askdirectory(initialdir=self.folder_buf, title=_("AD_TITLE1"))
        if new_folder:
            folder_fig = os.path.join(new_folder, "figures")
            if os.path.exists(folder_fig):
                self.folder1 = folder_fig
                self.folder1_label.config(text=_("LB_FOLDER1").format(os.path.basename(new_folder)))
                self.update_file_list()
                self.folder_buf = str(Path(new_folder).parent)
            else:
                MB.showerror(_("ERROR"), _("NO_FIG"))

    def open_folder2(self):
        new_folder = filedialog.askdirectory(initialdir=self.folder_buf, title=_("AD_TITLE2"))
        if new_folder:
            folder_fig = os.path.join(new_folder, "figures")
            if os.path.exists(folder_fig):
                self.folder2 = folder_fig
                self.folder2_label.config(text=_("LB_FOLDER2").format(os.path.basename(new_folder)))
                self.update_file_list()
                self.folder_buf = str(Path(new_folder).parent)
            else:
                MB.showerror(_("ERROR"), _("NO_FIG"))

    def update_file_list(self):
        self.file_listbox.delete(0, END)
        if self.folder1 and self.folder2:
            files1 = set(os.listdir(self.folder1))
            files2 = set(os.listdir(self.folder2))
            common_files = files1.union(files2)
            for file in sorted(common_files):
                if file.lower().endswith(('png', 'jpg', 'jpeg', 'gif')):  # 只列出图片文件
                    self.file_listbox.insert(END, file)

    def on_file_select(self, event):
        selected_index = self.file_listbox.curselection()
        if selected_index:
            file_name = self.file_listbox.get(selected_index)
            self.display_images(file_name)

    def resize(self):
        sz = (self.winfo_width() - 200, self.winfo_height() // 2 - 20)
        if self.original_image1 is not None:
            resized_image1 = self.original_image1.copy()
            resized_image1.thumbnail(sz)
            image1 = ImageTk.PhotoImage(resized_image1)

            self.image1_label.config(image=image1,text="")
            self.image1 = image1
        else:
            self.image1_label.config(image='',text=_("NO_IMAGE"))
            self.image1 = None

        if self.original_image2 is not None:
            resized_image2 = self.original_image2.copy()
            resized_image2.thumbnail(sz)
            image2 = ImageTk.PhotoImage(resized_image2)

            self.image2_label.config(image=image2,text="")
            self.image2 = image2
        else:
            self.image2_label.config(image='',text=_("NO_IMAGE"))
            self.image2 = None


    def on_resize(self, event):
        self.resize()
        
    def display_images(self, file_name:str):
        if self.folder1 is None: return
        img1_path = os.path.join(self.folder1, file_name)
        if self.folder2 is None: return
        img2_path = os.path.join(self.folder2, file_name)
        
        try:
            if os.path.exists(img1_path):
                self.original_image1 = Image.open(img1_path)
            else:
                self.original_image1 = None
            if os.path.exists(img2_path):
                self.original_image2 = Image.open(img2_path)
            else:
                self.original_image2 = None
        except Exception as e:
            MB.showerror(_("ERROR"), _("LOAD_FAILED").format(str(e)))
        
        self.resize()


import math
import os
import shutil
import sys
import numpy as np
import threading
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union, Set
from collections import defaultdict
from dataclasses import dataclass, field
from sklearn.cluster import KMeans
from sklearn.neighbors import KDTree
from collections import defaultdict
from .utils import DetectFiles, ReadXML
from ..locale import Lang


def _largeStackExec(func, *args):
    import sys
    threading.stack_size(67108864) #64MB
    sys.setrecursionlimit(10**6)
    th = threading.Thread(target=func, args=args)
    th.start()
    th.join()


class Node:
    def __init__(self, node_id:str, x:float, y:float):
        self.id = node_id
        self.x = x
        self.y = y
        self.incoming_edges:List[Edge] = []
        self.outgoing_edges:List[Edge] = []
    
    def get_coord(self) -> Tuple[float, float]:
        return (self.x, self.y)


@dataclass
class Edge:
    id: str
    from_node: Node
    to_node: Node
    length: float # in meters
    lanes: int = 1
    speed_limit: float = 13.89  # Default 50 km/h in m/s
    world_id: int = -1
    
@dataclass
class SubNet:
    nodes: Set[str] = field(default_factory=set)
    edges: Set[str] = field(default_factory=set)

class _TarjanSCC:
    def __init__(self, n:int, gl:Dict[int, List[int]]):
        self.__dfn: List[int] = [0] * n
        self.__low: List[int] = [0] * n
        self.__dfncnt = 0

        self.__scc: List[int] = [0] * n
        self.__sc = 0

        self.__stk: List[int] = []
        self.__vis: set[int] = set()
        self.__gl = gl

        self.scc = None
    
    def __tarjan(self, u: int):
        self.__dfncnt += 1
        self.__low[u] = self.__dfn[u] = self.__dfncnt
        self.__stk.append(u)
        self.__vis.add(u)
        for v in self.__gl[u]:
            if self.__dfn[v] == 0:
                self.__tarjan(v)
                self.__low[u] = min(self.__low[u], self.__low[v])
            elif v in self.__vis:
                self.__low[u] = min(self.__low[u], self.__dfn[v])
        if self.__low[u] == self.__dfn[u]:
            self.__sc += 1
            while True:
                v = self.__stk.pop()
                self.__vis.remove(v)
                self.__scc[v] = self.__sc
                if v == u:
                    break
    
    def calc_scc(self):
        """
        Calculate the strongly connected components (SCC) of the graph.
        Returns a list of SCC in a decreasing order by size, each element indicates the nodes in this SCC
        """
        n = len(self.__gl)
        for u in range(n):
            if self.__low[u] == 0:
                self.__tarjan(u)

        scc_dict: Dict[int, List[int]] = defaultdict(list)
        for i, x in enumerate(self.__scc):
            scc_dict[x].append(i)

        ret = list(scc_dict.values())
        ret.sort(key=lambda x: len(x), reverse=True)

        self.scc = ret


class RoadNet:
    VERSION = "1.0"
    def __init__(self):
        self.nodes:Dict[str, Node] = {}
        self.__nodeL:List[Node] = []
        self.edges:Dict[str, Edge] = {}
        self.__scc:List[SubNet] = []
        self.__kdt:Optional[KDTree] = None
        self._proj = None
        self.netOffset = (0.,0.)
        self.convBoundary = (0.,0.,0.,0.)
        self.origBoundary = (0.,0.,0.,0.)
        self.projParameter = "!"
    
    def calc_kdtree(self):
        """
        Calculate the KDTree of the road network nodes for fast nearest neighbor search.
        """
        self.__nodeL = list(self.nodes.values())
        coords = np.array([node.get_coord() for node in self.__nodeL])
        self.__kdt = KDTree(coords, metric='euclidean')

    @property
    def kdtree(self) -> KDTree:
        if self.__kdt is None:
            self.calc_kdtree()
        assert self.__kdt is not None
        return self.__kdt
    
    def check_scc_size(self, display:bool = True):
        '''Check if the size of the largest strongly connected component is large enough'''
        if len(self.scc[0].nodes) < len(self.nodes) * 0.8:
            if display: print(Lang.WARN_SCC_TOO_SMALL.format(len(self.scc[0].nodes), len(self.nodes)))
            return False
        return True
    
    def is_node_in_largest_scc(self, node_id:str) -> bool:
        """
        Check if a node is in the largest strongly connected component.
        """
        if len(self.__scc) == 0:
            self.calc_max_scc()
        return node_id in self.__scc[0].nodes
    
    def find_nearest_node(self, x:float, y:float) -> Node:
        """
        Find the nearest node to the given coordinates.
        """
        dist, idx = self.kdtree.query([[x, y]], k=1)
        return self.__nodeL[idx[0][0]]
    
    def find_nearest_node_with_distance(self, x:float, y:float) -> Tuple[float, Node]:
        """
        Find the nearest node to the given coordinates.
        """
        dist, idx = self.kdtree.query([[x, y]], k=1)
        return dist[0][0], self.__nodeL[idx[0][0]]
    
    def calc_max_scc(self):
        """
        Calculate the largest strongly connected component (SCC) of the road network.
        Returns a set of node IDs in the largest SCC.
        """
        nodes = list(self.nodes.values())
        nmp = {node.id: i for i, node in enumerate(nodes)}
        tscc = _TarjanSCC(n=self.node_count, gl={
            i: [nmp[e.to_node.id] for e in nodes[i].outgoing_edges] 
            for i in range(len(self.nodes))
        })
        _largeStackExec(tscc.calc_scc)
        
        assert tscc.scc is not None

        sccidx_of_nodes = {}
        for i, node_ids in enumerate(tscc.scc):
            for nid in node_ids:
                sccidx_of_nodes[nid] = i
        
        scc_tmp = {
            i: SubNet(nodes = {nodes[j].id for j in node_ids}) 
            for i, node_ids in enumerate(tscc.scc)
        }
        for edge in self.edges.values():
            if sccidx_of_nodes[nmp[edge.from_node.id]] == sccidx_of_nodes[nmp[edge.to_node.id]]:
                scc_tmp[sccidx_of_nodes[nmp[edge.from_node.id]]].edges.add(edge.id)
            
        self.__scc = list(scc_tmp.values())
        return self.__scc
    
    @property
    def scc(self) -> List[SubNet]:
        if len(self.__scc) == 0:
            self.calc_max_scc()
        return self.__scc

    def add_node(self, node_id:str, x:int, y:int) -> Node:
        self.__scc = []
        self.__kdt = None
        if node_id in self.nodes:
            raise ValueError(f"Node {node_id} already exists.")
        node = Node(node_id, x, y)
        self.nodes[node_id] = node
        return node
    
    def get_node(self, node_id:str) -> Node:
        return self.nodes[node_id]
    
    def rename_node(self, old_id:str, new_id:str):
        if new_id in self.nodes:
            raise ValueError(Lang.NODE_EXISTS.format(new_id))
        node = self.nodes.pop(old_id)
        for edge in node.incoming_edges:
            edge.to_node = node
        for edge in node.outgoing_edges:
            edge.from_node = node
        node.id = new_id
        self.nodes[new_id] = node
    
    def remove_node(self, node_id:str):
        self.__scc = []
        self.__kdt = None
        if not node_id in self.nodes:
            raise ValueError(Lang.NODE_NOT_FOUND.format(node_id))
        node = self.nodes.pop(node_id)
        enames = list(self.edges.keys())
        for e in enames:
            edge = self.edges[e]
            if edge.from_node == node or edge.to_node == node:
                self.remove_edge(e)
    
    def update_node(self, old_node_id:str, new_node_id:str):
        if not old_node_id in self.nodes:
            raise ValueError(Lang.NODE_NOT_FOUND.format(old_node_id))
        if new_node_id in self.nodes:
            raise ValueError(Lang.NODE_EXISTS.format(new_node_id))
        node = self.nodes.pop(old_node_id)
        node.id = new_node_id
        self.nodes[new_node_id] = node
        
    @property
    def node_ids(self):
        return list(self.nodes.keys())
    
    @property
    def node_count(self):
        return len(self.nodes)
    
    def add_edge(self, edge_id:str, from_node:Union[str, Node], to_node:Union[str, Node], 
            length_m:float, lanes:int, speed_limit:float, world_id:int = -1) -> Edge:
        self.__scc = []
        if edge_id in self.edges:
            raise ValueError(Lang.EDGE_EXISTS.format(edge_id))
        if isinstance(from_node, str): from_node = self.get_node(from_node)
        if isinstance(to_node, str): to_node = self.get_node(to_node)
        edge = Edge(edge_id, from_node, to_node, length_m, lanes, speed_limit, world_id)
        self.edges[edge_id] = edge
        from_node.outgoing_edges.append(edge)
        to_node.incoming_edges.append(edge)
        return edge
    
    def get_edge(self, edge_id:str) -> Edge:
        return self.edges[edge_id]
    
    def rename_edge(self, old_id:str, new_id:str):
        if new_id in self.edges:
            raise ValueError(Lang.EDGE_EXISTS.format(new_id))
        edge = self.edges.pop(old_id)
        edge.id = new_id
        self.edges[new_id] = edge
    
    def remove_edge(self, edge_id:str):
        self.__scc = []
        if not edge_id in self.edges:
            raise ValueError(Lang.EDGE_NOT_FOUND.format(edge_id))
        self.edges.pop(edge_id)
    
    def update_edge(self, edge_id:str, new_from_node_id:str, new_to_node_id:str):
        if not edge_id in self.edges:
            raise ValueError(Lang.EDGE_NOT_FOUND.format(edge_id))
        if new_from_node_id not in self.nodes:
            raise ValueError(Lang.NODE_NOT_FOUND.format(new_from_node_id))
        if new_to_node_id not in self.nodes:
            raise ValueError(Lang.NODE_NOT_FOUND.format(new_to_node_id))
        e = self.edges[edge_id]
        new_from_node = self.nodes[new_from_node_id]
        if e.from_node != new_from_node:
            e.from_node.outgoing_edges.remove(e)
            new_from_node.outgoing_edges.append(e)
            e.from_node = new_from_node
        new_to_node = self.nodes[new_to_node_id]
        if e.to_node != new_to_node:
            e.to_node.incoming_edges.remove(e)
            new_to_node.incoming_edges.append(e)
            e.to_node = new_to_node
    
    def get_offset_shape(self, edge_id:str):
        e = self.get_edge(edge_id)
        start_x, start_y = e.from_node.get_coord()
        end_x, end_y = e.to_node.get_coord()

        # Direction vector
        dx = end_x - start_x
        dy = end_y - start_y
        length = math.hypot(dx, dy)
        if length == 0:
            dir_dx, dir_dy = 1, 0
        else:
            dir_dx = dx / length
            dir_dy = dy / length

        # Perpendicular vector (right side)
        right_dx = dir_dy
        right_dy = -dir_dx

        # Offset along the right side
        minx, miny, maxx, maxy = self.getBoundary()
        offset_px = math.hypot(maxx - minx, maxy - miny) * 1e-3
        start_x += right_dx * offset_px
        start_y += right_dy * offset_px
        end_x += right_dx * offset_px
        end_y += right_dy * offset_px

        return ((start_x, start_y), (end_x, end_y))
    
    @property
    def edge_ids(self):
        return list(self.edges.keys())
    
    @property
    def edge_count(self):
        return len(self.edges)
    
    @staticmethod
    def load_raw(fname:str):
        ret = RoadNet()
        root = ReadXML(fname)
        
        if root is None:
            raise RuntimeError(f"Invalid xml file: {fname}")
        for node in root.findall("node"):
            ret.add_node(
                node_id = node.attrib["id"],
                x = int(float(node.attrib.get("x", "0"))),
                y = int(float(node.attrib.get("y", "0")))
            )
        for edge in root.findall("edge"):
            ret.add_edge(
                edge_id = edge.attrib["id"],
                from_node = edge.attrib["from"],
                to_node = edge.attrib["to"],
                length_m = float(edge.attrib["length"]),
                lanes = int(edge.attrib.get("lanes", "1")),
                speed_limit = float(edge.attrib.get("speed", "13.89")),  # Default 50 km/h in m/s
                world_id = int(edge.attrib.get("world_id", "-1"))
            )
        location = root.find("location")
        if location is not None:
            ret.netOffset = tuple(map(float, location.attrib.get("netOffset", "0,0").split(",")))
            ret.convBoundary = tuple(map(float, location.attrib.get("convBoundary", "0,0,0,0").split(",")))
            ret.origBoundary = tuple(map(float, location.attrib.get("origBoundary", "0,0,0,0").split(",")))
            ret.projParameter = location.attrib.get("projParameter", "!")
        return ret
    
    @staticmethod
    def load_sumo(fname:str, only_passenger:bool=True):
        ret = RoadNet()
        from sumolib.net import readNet, Net
        r: Net = readNet(fname)
        assert isinstance(r, Net), Lang.INVALID_SUMO_NETWORK.format(fname)
        for node in r.getNodes():
            ret.add_node(
                node_id = node.getID(),
                x = int(node.getCoord()[0]),
                y = int(node.getCoord()[1])
            )
        for edge in r.getEdges():
            if only_passenger and not edge.allows("passenger"): continue
            ret.add_edge(
                edge_id = edge.getID(),
                from_node = edge.getFromNode().getID(),
                to_node = edge.getToNode().getID(),
                length_m = edge.getLength(),
                lanes = edge.getLaneNumber(),
                speed_limit = edge.getSpeed(),
                world_id = -1
            )
        if len(r._location) > 0:
            ret.netOffset = tuple(r.getLocationOffset())
            ret.convBoundary = tuple(r.getBoundary())
            ret.origBoundary = tuple(map(float, r._location["origBoundary"].split(',')))
            ret.projParameter = r._location["projParameter"]
        return ret
    
    @staticmethod
    def load(fname:str, fmt:str="auto"):
        if fmt == "raw":
            return RoadNet.load_raw(fname)
        elif fmt == "sumo":
            return RoadNet.load_sumo(fname)
        elif fmt == "auto":
            try:
                return RoadNet.load_sumo(fname)
            except:
                return RoadNet.load_raw(fname)
        else:
            raise ValueError(Lang.UNKNOWN_NET_FORMAT.format(fmt))
    
    def save(self, fname:str):
        if fname.lower().endswith(".gz"):
            import gzip
            f = gzip.open(fname, "wt")
        else:
            f = open(fname, "w")
        f.write(f'<roadnet v2simfmtver="{RoadNet.VERSION}">\n')
        def _w(x): return ','.join(map(str, x))
        if self.hasGeoProj():
            f.write(f'  <location netOffset="{_w(self.netOffset)}" convBoundary="{_w(self.convBoundary)}" origBoundary="{_w(self.origBoundary)}" projParameter="{self.projParameter}"/>')
        for node in self.nodes.values():
            f.write(f'  <node id="{node.id}" x="{node.x}" y="{node.y}"/>\n')
        for edge in self.edges.values():
            f.write(f'  <edge id="{edge.id}" from="{edge.from_node.id}" to="{edge.to_node.id}" length="{edge.length}" lanes="{edge.lanes}" speed="{edge.speed_limit}" world_id="{edge.world_id}"/>\n')
        f.write("</roadnet>\n")
        f.close()
    
    def create_world(self, **kwargs):
        for edge in self.edges.values():
            world_id = edge.world_id
            break
        if all(edge.world_id == world_id for edge in self.edges.values()):
            return self.create_singleworld(**kwargs)
        else:
            if not hasattr(sys, "_is_gil_enabled") or sys._is_gil_enabled(): # type: ignore
                print(Lang.GIL_NOT_DISABLED)
                return self.create_singleworld(**kwargs)
            return self.create_paraworlds(**kwargs)

    def create_singleworld(self, **kwargs):
        from .uxsim import World
        from .paraworlds import SingleWorld
        from .routing import Graph

        kwargs.pop("name", None)
        kwargs.pop("save_mode", None)
        world = World(name="0", save_mode=0, **kwargs)
           
        gl:Graph = {nid: [] for nid in self.nodes}
        for edge in self.edges.values():
            fr = edge.from_node.id
            to = edge.to_node.id
            if fr not in world.NODES_NAME_DICT:
                world.addNode(name = fr, x = edge.from_node.x, y = edge.from_node.y)
            if to not in world.NODES_NAME_DICT:
                world.addNode(name = to, x = edge.to_node.x, y = edge.to_node.y)
            link = world.addLink(name = edge.id, start_node = edge.from_node.id, end_node = edge.to_node.id,
                length = edge.length, free_flow_speed = edge.speed_limit, number_of_lanes = edge.lanes)
            gl[fr].append((to, link))
        
        return SingleWorld(world, gl)

    def create_paraworlds(self, **kwargs):
        from .uxsim import World
        from .paraworlds import ParaWorlds
        from .routing import Graph

        kwargs.pop("name", None)
        kwargs.pop("print_mode", None)
        kwargs.pop("save_mode", None)

        # Check if all edges do not specify world_id
        if all(edge.world_id == -1 for edge in self.edges.values()):
            for edge in self.edges.values():
                edge.world_id = 0
        
        if any(edge.world_id == -1 for edge in self.edges.values()):
            raise RuntimeError("Some edges do not specify world_id while others do. Please specify world_id for all edges or none.")
        
        worlds:Dict[int, World] = {}
        gl:Graph = {nid: [] for nid in self.nodes}
        for edge in self.edges.values():
            wid = edge.world_id
            if wid not in worlds:
                worlds[wid] = World(name=str(wid), print_mode=0, save_mode=0, **kwargs)
            W = worlds[wid]
            fr = edge.from_node.id
            to = edge.to_node.id
            if fr not in W.NODES_NAME_DICT:
                W.addNode(name = fr, x = edge.from_node.x, y = edge.from_node.y)
            if to not in W.NODES_NAME_DICT:
                W.addNode(name = to, x = edge.to_node.x, y = edge.to_node.y)
            link = W.addLink(name = edge.id, start_node = fr, end_node = to,
                length = edge.length, free_flow_speed = edge.speed_limit, number_of_lanes = edge.lanes)
            gl[fr].append((to, link))
        
        return ParaWorlds(worlds, gl)

    def hasGeoProj(self):
        return self.projParameter != "!"

    def getGeoProj(self):
        if not self.hasGeoProj():
            raise RuntimeError(Lang.NO_GEO_PROJ)
        if self._proj is None:
            import pyproj
            try:
                self._proj = pyproj.Proj(projparams=self.projParameter)
            except RuntimeError:
                if hasattr(pyproj.datadir, 'set_data_dir'): # type: ignore
                    pyproj.datadir.set_data_dir('/usr/share/proj') # type: ignore
                    self._proj = pyproj.Proj(projparams=self.projParameter)
                raise
        return self._proj

    def getLocationOffset(self):
        """ offset to be added after converting from geo-coordinates to UTM"""
        return self.netOffset

    def getBoundary(self):
        """ return xmin,ymin,xmax,ymax network coordinates"""
        if self.convBoundary == (0.,0.,0.,0.):
            all_x = [node.x for node in self.nodes.values()]
            all_y = [node.y for node in self.nodes.values()]
            x_max, x_min = max(all_x), min(all_x)
            y_max, y_min = max(all_y), min(all_y)
            return x_min, y_min, x_max, y_max
        return self.convBoundary

    def convertLonLat2XY(self, lon, lat, rawUTM=False):
        x, y = self.getGeoProj()(lon, lat)
        if rawUTM:
            return x, y
        else:
            x_off, y_off = self.getLocationOffset()
            return x + x_off, y + y_off

    def convertXY2LonLat(self, x, y, rawUTM=False):
        if not rawUTM:
            x_off, y_off = self.getLocationOffset()
            x -= x_off
            y -= y_off
        return self.getGeoProj()(x, y, inverse=True)

    def partition_roadnet(self, num_partitions: int) -> None:
        edge_groups = self._group_reverse_edges()

        partition_assignment = self._geographic_clustering(edge_groups, num_partitions)
        
        for edge_id, partition_id in partition_assignment.items():
            self.edges[edge_id].world_id = partition_id

    def _group_reverse_edges(self) -> List[Set[str]]:
        edge_lookup = {}
        for edge_id, edge in self.edges.items():
            key = (edge.from_node.id, edge.to_node.id)
            edge_lookup[key] = edge_id
        
        visited = set()
        edge_groups = []
        
        for edge_id, edge in self.edges.items():
            if edge_id in visited:
                continue
                
            current_group = {edge_id}
            visited.add(edge_id)
            
            reverse_key = (edge.to_node.id, edge.from_node.id)
            if reverse_key in edge_lookup:
                reverse_edge_id = edge_lookup[reverse_key]
                current_group.add(reverse_edge_id)
                visited.add(reverse_edge_id)
            
            edge_groups.append(current_group)
        
        return edge_groups

    def _geographic_clustering(self, edge_groups: List[Set[str]], num_partitions: int) -> Dict[str, int]:
        group_features = []
        group_edges = []  # 记录每个组包含的边ID
        
        for edge_group in edge_groups:
            # 计算该组所有边的平均坐标
            coords = []
            for edge_id in edge_group:
                edge = self.edges[edge_id]
                # 使用边的中点作为坐标
                mid_x = (edge.from_node.x + edge.to_node.x) / 2
                mid_y = (edge.from_node.y + edge.to_node.y) / 2
                coords.append([mid_x, mid_y])
            
            avg_coord = np.mean(coords, axis=0)
            group_features.append(avg_coord)
            group_edges.append(list(edge_group))
        
        # 使用K-means进行地理聚类
        if len(group_features) <= num_partitions:
            # 如果组数少于分区数，直接分配
            assignment = {}
            for i, edge_group in enumerate(group_edges):
                partition_id = i % num_partitions
                for edge_id in edge_group:
                    assignment[edge_id] = partition_id
        else:
            # 使用K-means聚类
            kmeans = KMeans(n_clusters=num_partitions, random_state=42, n_init=10)
            group_labels = kmeans.fit_predict(group_features)
            
            assignment = {}
            for i, (edge_group, label) in enumerate(zip(group_edges, group_labels)):
                for edge_id in edge_group:
                    assignment[edge_id] = label
        
        return assignment

    @property
    def world_count(self):
        world_ids = set(edge.world_id for edge in self.edges.values())
        return len(world_ids) if world_ids else 1
    
    def create_color_map(self):
        import matplotlib.pyplot as plt
        import matplotlib.colors as mcolors
        # 获取所有不同的world_id
        world_ids = set(edge.world_id for edge in self.edges.values())
        num_partitions = len(world_ids) if world_ids else 1
        
        # 生成颜色映射
        if num_partitions <= 10:
            colors = list(mcolors.TABLEAU_COLORS.values())[:num_partitions]
        else:
            cmap = plt.cm.get_cmap('tab20', num_partitions)
            colors = [cmap(i) for i in range(num_partitions)]
        colors = [mcolors.to_hex(c) for c in colors]
        
        return {wid: colors[i] for i, wid in enumerate(sorted(world_ids))}

    def plot_roadnet(self, 
            figsize: Tuple[int, int] = (12, 10),
            node_size: int = 50,
            show_nodes: bool = True,
            show_labels: bool = False,
            title: str = "Road Network Partitioning") -> None:
        import matplotlib.pyplot as plt
        from matplotlib.lines import Line2D

        fig, ax = plt.subplots(figsize=figsize)
        
        # 获取所有node的坐标范围
        x_min, y_min, x_max, y_max = self.getBoundary()

        color_map = self.create_color_map()
        
        # 首先绘制边
        for edge in self.edges.values():
            from_node = edge.from_node
            to_node = edge.to_node
            
            start_x, start_y = from_node.x, from_node.y
            end_x, end_y = to_node.x, to_node.y

            # 计算方向向量
            dx = end_x - start_x
            dy = end_y - start_y
            length = np.sqrt(dx**2 + dy**2)
            if length == 0:
                dir_dx, dir_dy = 1, 0
            else:
                dir_dx = dx / length
                dir_dy = dy / length

            # 计算垂直向量（右侧方向）
            right_dx = dir_dy
            right_dy = -dir_dx

            # 沿右侧偏移
            offset_px = math.hypot(x_max - x_min, y_max - y_min) * 1e-3
            start_x += right_dx * offset_px
            start_y += right_dy * offset_px
            end_x += right_dx * offset_px
            end_y += right_dy * offset_px
            
            # 获取颜色
            color = color_map.get(edge.world_id, 'gray')
            
            lw = min(2, offset_px * 0.02)
            # 绘制边
            ax.plot([start_x, end_x], 
                    [start_y, end_y], 
                    color=color, 
                    linewidth=lw,
                    linestyle='-',
                    alpha=0.8)
            
            # 绘制箭头
            mid_x = (start_x + end_x) / 2
            mid_y = (start_y + end_y) / 2

            arrow_dx = dx * 0.2
            arrow_dy = dy * 0.2
            
            ax.arrow(mid_x - arrow_dx/2, mid_y - arrow_dy/2, 
                    arrow_dx, arrow_dy, width=lw,
                    head_width=lw*5, head_length=lw*3,
                    fc=color, ec=color, alpha=0.8)
        
        # 绘制节点
        if show_nodes:
            node_x = [node.x for node in self.nodes.values()]
            node_y = [node.y for node in self.nodes.values()]
            node_ids = [node.id for node in self.nodes.values()]
            
            ax.scatter(node_x, node_y, s=node_size, c='black', alpha=0.7, zorder=5)
            
            if show_labels:
                for i, node_id in enumerate(node_ids):
                    ax.annotate(node_id, (node_x[i], node_y[i]), 
                            xytext=(5, 5), textcoords='offset points',
                            fontsize=8, alpha=0.8)
        
        ax.set_xlabel('X Coordinate')
        ax.set_ylabel('Y Coordinate')
        ax.set_title(title)
        ax.grid(True, alpha=0.3)
        
        # 添加图例
        legend_elements = []
        for wid, color in color_map.items():
            legend_elements.append(Line2D([0], [0], color=color, lw=4, label=f'Partition {wid}'))
        ax.legend(handles=legend_elements, loc='best')
        
        plt.tight_layout()
        plt.savefig("roadnet_partitioned.png", dpi=300)
    
    def remove_items_outside_max_scc(self):
        max_scc = self.scc[0]
        to_remove = []
        for n in self.nodes:
            if n not in max_scc.nodes:
                to_remove.append(n)
        for n in to_remove:
            self.remove_node(n)

        to_remove = []
        for e in self.edges:
            if e not in max_scc.edges:
                to_remove.append(e)
        for e in to_remove:
            self.remove_edge(e)


def ConvertCase(input_dir:str, output_dir:str, part_cnt:int, auto_partition:bool,
            non_passenger_links:bool, non_scc_links:bool):
    """Convert case files in input directory and save to output directory.
    
    Args:
        input_dir: Input directory path.
        output_dir: Output directory path.
        pcnt: Partition count.
        auto_part: Whether to auto determine partition count.
        non_passenger: Whether to include non-passenger links.
        non_scc: Whether to include links and edges not in the largest SCC.
    """
    files = DetectFiles(input_dir)
    converted = False
    out_dir = Path(output_dir)
    out_dir.mkdir(parents=True, exist_ok=True)
    if files.net:
        print("Found SUMO network file:", files.net)
        r = RoadNet.load_sumo(files.net, only_passenger=not non_passenger_links)
        if not non_scc_links:
            print("Extracting largest strongly connected component...")
            r.remove_items_outside_max_scc()
        if auto_partition:
            part_cnt = min(32, os.cpu_count() or 1, r.node_count // 40)
            print(f"Auto partition count determined: {part_cnt}")
        if part_cnt > 1:
            print(f"Partitioning network into {part_cnt} parts...")
            r.partition_roadnet(part_cnt)
        r.save(str(out_dir / Path(files.net).name))
        converted = True
    if files.poly:
        print("Found POLY file:", files.poly)
        shutil.copy(files.poly, str(out_dir / Path(files.poly).name))
        converted = True
    if files.cscsv:
        print("Found charging station CSV file:", files.cscsv)
        shutil.copy(files.cscsv, str(out_dir / Path(files.cscsv).name))
        converted = True
    if files.osm:
        print("Found OSM file:", files.osm)
        shutil.copy(files.osm, str(out_dir / Path(files.osm).name))
        converted = True
    if files.grid:
        print("Found power grid file:", files.grid)
        shutil.copy(files.grid, str(out_dir / Path(files.grid).name))
        converted = True
    if files.py:
        print("Found vehicle Python file:", files.py)
        shutil.copy(files.py, str(out_dir / Path(files.py).name))
        converted = True
    if files.pref:
        print("Found vehicle preference file:", files.pref)
        shutil.copy(files.pref, str(out_dir / Path(files.pref).name))
        converted = True
    if files.plg:
        print("Found plugin file:", files.plg)
        shutil.copy(files.plg, str(out_dir / Path(files.plg).name))
        converted = True
    if not converted:
        print("No supported files found in the input directory.")
    return converted


__all__ = ["Node", "Edge", "SubNet", "RoadNet", "ConvertCase"]


from v2simux_gui.common import *

from v2simux import StaPool
from .pareditor import ParamsEditor
from ..mainbox.controls import LogItemPad
from .utils import *


class LoadGroupBox(Toplevel):
    def __init__(self, parent, folder:str):
        super().__init__(parent)
        self.params = {}
        self.folder = folder
        self.results:Optional[list[tuple[str,str]]] = None
        self.title(_L("LOAD_GROUP_TITLE"))
        self.geometry('600x300')
        self.lb = Label(self, text=_L("FOLDER_NAME").format(self.folder))
        self.lb.pack(padx=3, pady=3)
        self.fr = Frame(self)
        self.fr.pack(padx=3, pady=3)
        self.lb_p = Label(self.fr, text=_L("OTHER_PARAMS"))
        self.lb_p.grid(row=0, column=0, padx=3, pady=3)
        self.fr2 = Frame(self.fr)
        self.fr2.grid(row=0, column=1, padx=3, pady=3)
        self.lb_pv = Label(self.fr2, text=str(self.params))
        self.lb_pv.pack(padx=3, anchor=W, side=LEFT)
        self.en_p = Button(self.fr2, command=self.edit_params, text="...",width=3)
        self.en_p.pack(padx=3, anchor=W, side=LEFT)
        self.lb_m = Label(self.fr, text=_L("MODE_ITEM"))
        self.lb_m.grid(row=1, column=0, padx=3, pady=3)
        self.cb = Combobox(self.fr)
        self.cb.grid(row=1, column=1, padx=3, pady=3)
        self.cb["values"] = [ITEM_NONE, "scs_slots", "fcs_slots", "start_time", "end_time", "traffic_step"]
        self.cb.current(0)
        self.lb_s = Label(self.fr, text=_L("START_VALUE"))
        self.lb_s.grid(row=2, column=0, padx=3, pady=3)
        self.en_s = Entry(self.fr)
        self.en_s.grid(row=2, column=1, padx=3, pady=3)
        self.lb_e = Label(self.fr, text=_L("END_VALUE"))
        self.lb_e.grid(row=3, column=0, padx=3, pady=3)
        self.en_e = Entry(self.fr)
        self.en_e.grid(row=3, column=1, padx=3, pady=3)
        self.lb_t = Label(self.fr, text=_L("STEP_VALUE"))
        self.lb_t.grid(row=4, column=0, padx=3, pady=3)
        self.en_t = Entry(self.fr)
        self.en_t.grid(row=4, column=1, padx=3, pady=3)
        self.lip = LogItemPad(self, _L["SIM_STAT"], StaPool())
        self.lip["ev"]=False
        self.lip.pack(padx=3, pady=3)
        self.btn = Button(self, text=_L("LGB_WORK"), command=self.work)
        self.btn.pack(padx=3, pady=3)
    
    def edit_params(self):
        pe = ParamsEditor(self.params)
        pe.wait_window()
        self.params = pe.data
        self.lb_pv["text"] = str(self.params)
    
    def work(self):
        self.results = []
        mode = self.cb.get()
        if mode == ITEM_NONE:
            self.results.append(('{}', ''))
        else:
            ms = ''.join(x[0] for x in mode.split('_'))
            try:
                start = int(self.en_s.get())
                end = int(self.en_e.get())
                step = int(self.en_t.get())
            except ValueError:
                MB.showerror(_L("ERROR"), _L("INVALID_VALUE"))
                self.focus()
                return
            for i in range(start, end, step):
                self.results.append(('{'+f"'{mode}':{i}"+'}', f"{ms}_{i}"))
        self.destroy()

__all__ = ["LoadGroupBox"]


from v2simux_gui.com_no_vx import *
from v2simux_gui.langhelper import add_lang_menu

import os
import gzip
import traceback
from v2simux import ReadOnlyStatistics, AdvancedPlot, SAVED_STATE_FOLDER, TRAFFIC_INST_FILE_NAME, TrafficInst
from PIL import Image, ImageTk
from .srd import SelectResultsDialog
from .plotpage import PlotPage, AVAILABLE_ITEMS, AVAILABLE_ITEMS2
from .trips import TripsFrame
from .gridpage import GridFrame
from .statepage import StateFrame


_L = LangLib.Load(__file__)

ITEM_ALL = _L["ITEM_ALL"]
ITEM_SUM = _L["ITEM_SUM"]
ITEM_ALL_G = "<All common generators>"
ITEM_ALL_V2G = "<All V2G stations>"
ITEM_LOADING = "Loading..."


class ViewerBox(Tk):
    _sta:ReadOnlyStatistics
    _npl:AdvancedPlot

    def __init__(self, project_folder:Optional[str] = None):
        super().__init__()
        self.title(_L["TITLE"])
        self.geometry("1024x840")
        self.original_image = None
        
        self.menu = Menu(self)
        self.config(menu=self.menu)
        self.filemenu = Menu(self.menu, tearoff=False)
        self.menu.add_cascade(label=_L["MENU_FILE"], menu=self.filemenu)
        self.filemenu.add_command(label=_L["MENU_OPEN"], command=self.open)
        self.filemenu.add_separator()
        self.filemenu.add_command(label=_L["MENU_EXIT"], command=self.destroy)
        add_lang_menu(self.menu)

        self.tab = Notebook(self)
        self.tab.pack(expand=True,fill='both',padx=20,pady=3)
        self.tab_curve = Frame(self.tab)
        self.tab.add(self.tab_curve,text=_L["TAB_CURVE"])
        self.tab_grid = GridFrame(self.tab, self)
        self.tab.add(self.tab_grid,text=_L["TAB_GRID"])
        self.tab_trip = TripsFrame(self.tab)
        self.tab.add(self.tab_trip,text=_L["TAB_TRIP"], sticky='nsew')
        self.tab_state = StateFrame(self.tab)
        self.tab.add(self.tab_state,text=_L["TAB_STATE"], sticky='nsew')

        self.fr_pic = Frame(self.tab_curve)
        self.fr_pic.pack(side="top",fill=BOTH, expand=False)
        self.lb_pic = Label(self.fr_pic,text=_L["NO_IMAGE"])
        self.lb_pic.pack(side="left",fill=BOTH, expand=True,anchor='w')
        self.pic_list = Listbox(self.fr_pic)
        self.pic_list.pack(side="right",fill=Y,anchor='e')
        self.pic_list.bind("<<ListboxSelect>>", self.on_file_select)

        self.fr_draw = Frame(self.tab_curve)
        self.fr_draw.pack(side="bottom",fill=BOTH, expand=True)
        self._ppc = Canvas(self.fr_draw)
        self._ppc.pack(side="left", fill="both", expand=True)
        self.scrollbar = Scrollbar(self.fr_draw, orient="vertical", command=self._ppc.yview)
        self.scrollbar.pack(side="right", fill="y")
        self._pp = PlotPage(self._ppc)
        self._pp.bind("<Configure>", lambda e: self._ppc.configure(scrollregion=self._ppc.bbox("all")))
        self.btn_draw = Button(self._pp.panel_time, text=_L["BTN_PLOT"], command=self.plotSelected)
        self.btn_draw.pack(side='right')
        self._ppc.create_window((0, 0), window=self._pp, anchor="nw")
        self._ppc.configure(yscrollcommand=self.scrollbar.set)

        self._sbar=Label(self,text=_L["STA_READY"])
        self._sbar.pack(side='bottom',anchor='w',padx=3,pady=3)
        
        self._ava ={
            "fcs": False,
            "scs": False, 
            "ev": False, 
            "gen": False, 
            "bus": False, 
            "line": False, 
            "pvw": False, 
            "ess": False,
        }
        self._Q = EventQueue(self)
        self._Q.register("exit", self.quit)
        self._Q.register("loaded", self.on_loaded)
        self._Q.register("state_loaded", self.on_state_loaded)
        self._Q.register("plot_done", self.on_plot_done)
        self._Q.do()

        self.disable_all()
        self.bind("<Configure>", self.on_resize)
        self.resize_timer = None

        if project_folder is not None and project_folder != "":
            self.__proj_folder = project_folder
            self.bind("<Visibility>", self.__on_win_loaded)

    def __on_win_loaded(self, event):
        self.unbind("<Visibility>")
        self.load_results(self.__proj_folder)
    
    def display_images(self, file_name:str):
        if self.folder is None: return
        img1_path = os.path.join(self.folder, file_name)
        
        try:
            if os.path.exists(img1_path):
                self.original_image = Image.open(img1_path)
            else:
                self.original_image = None
        except Exception as e:
            MB.showerror(_L["ERROR"], _L["LOAD_FAILED"].format(str(e)))
        
        self.resize()
    
    def resize(self):
        sz = (self.winfo_width() - 200, self.winfo_height() // 2 - 20)
        if self.original_image is not None:
            resized_image = self.original_image.copy()
            resized_image.thumbnail(sz)
            image = ImageTk.PhotoImage(resized_image)

            self.lb_pic.config(image=image,text="")
            self.image = image
        else:
            self.lb_pic.config(image='',text=_L["NO_IMAGE"])
            self.image = None

    def on_resize(self, event):
        if self.resize_timer is not None:
            self.after_cancel(self.resize_timer)
        self.resize_timer = self.after(100, self.resize_end)
    
    def resize_end(self):
        self.resize()
    
    def on_file_select(self, event):
        selected_index = self.pic_list.curselection()
        if selected_index:
            file_name = self.pic_list.get(selected_index)
            self.display_images(file_name)
       
    def disable_all(self):
        self._pp.disable()
        self.btn_draw['state']=DISABLED

    def enable_all(self):
        self._pp.enable([p for p, ok in self._ava.items() if ok])
        self.btn_draw['state']=NORMAL
    
    def set_status(self,text:str):
        self._sbar.configure(text=text)

    def update_file_list(self):
        self.pic_list.delete(0, END)
        self.original_image = None
        self.lb_pic.config(image='',text=_L["NO_IMAGE"])
        self.image = None
        if self.folder and os.path.exists(self.folder):
            files = set(os.listdir(self.folder))
            for file in sorted(files):
                if file.lower().endswith(('png', 'jpg', 'jpeg', 'gif')):  # 只列出图片文件
                    self.pic_list.insert(END, file)

    def on_state_loaded(self, par):
        self.tab_state.setStateInst(par)
    
    def on_loaded(self, sta:ReadOnlyStatistics, npl:AdvancedPlot):
        assert isinstance(sta, ReadOnlyStatistics)
        assert isinstance(npl, AdvancedPlot)
        self._sta = sta
        self.tab_grid.setSta(sta)
        self._npl = npl
        for x in AVAILABLE_ITEMS:
            self._ava[x] = getattr(self._sta, f"has_{x.upper()}")()
        if self._sta.has_FCS():
            self._pp.fcs_pad.setValues([ITEM_SUM, ITEM_ALL] + self._sta.FCS_head)
            self.tab_state.cb_fcs_query['values'] = self._sta.FCS_head
            if self._sta.FCS_head:
                self.tab_state.cb_fcs_query.set(self._sta.FCS_head[0])
        if self._sta.has_SCS():
            self._pp.scs_pad.setValues([ITEM_SUM, ITEM_ALL] + self._sta.SCS_head)
            self.tab_state.cb_scs_query['values'] = self._sta.SCS_head
            if self._sta.SCS_head:
                self.tab_state.cb_scs_query.set(self._sta.SCS_head[0])
        if self._sta.has_GEN():
            self._pp.gen_pad.setValues([ITEM_ALL_G,ITEM_ALL_V2G,ITEM_ALL] + self._sta.gen_head)
        if self._sta.has_BUS():
            self._pp.bus_pad.setValues([ITEM_ALL] + self._sta.bus_head)
        if self._sta.has_LINE():
            self._pp.line_pad.setValues([ITEM_ALL] + self._sta.line_head)
        if self._sta.has_PVW():
            self._pp.pvw_pad.setValues([ITEM_ALL] + self._sta.pvw_head)
        if self._sta.has_ESS():
            self._pp.ess_pad.setValues([ITEM_ALL] + self._sta.ess_head)
        self.update_file_list()
        self.set_status(_L["STA_READY"])
        self.enable_all()
        
    def on_error(self, par):
        MB.showerror(_L["ERROR"], par[0])
        self.set_status(par[0])
        self.enable_all()
    
    def on_plot_done(self, ex:Optional[Exception] = None):
        if ex is None:
            self.update_file_list()
            self.set_status(_L["STA_READY"])
            self.enable_all()
        else:
            self.on_error((str(ex),))
    
    def askdir(self):
        p = Path(os.getcwd()) / "cases"
        p.mkdir(parents=True,exist_ok=True)
        return filedialog.askdirectory(
            title=_L["TITLE_SEL_FOLDER"],
            initialdir=str(p),
            mustexist=True,
        )
    
    def open(self):
        res_path = self.askdir()
        if res_path == "": return
        self.load_results(res_path)
    
    def load_results(self, folder: str):
        # Check folder existence
        first = True
        res_path = folder
        while True:
            res_path = Path(res_path)
            if res_path.exists():
                break
            else: 
                if not first: MB.showerror(_L["ERROR"], "Folder not found!")
            first = False
            res_path = self.askdir()
            if res_path == "":
                self._Q.trigger("exit")
                return
        
        # Check cproc.clog existence
        cproc = res_path / "cproc.clog"
        if cproc.exists():
            self.tab_trip.load(str(cproc))
        else:
            res_path_list = []
            for dir_ in res_path.iterdir():
                if dir_.is_dir() and dir_.name.lower().startswith("results") and (dir_ / "cproc.clog").exists():
                    res_path_list.append(dir_)
            if len(res_path_list) == 0:
                MB.showerror(_L["ERROR"], _L["NO_CPROC"])
                return
            elif len(res_path_list) == 1:
                res_path = res_path_list[0]
            else:
                self.disable_all()
                dsa = SelectResultsDialog(res_path_list)
                dsa.lift(self)
                self.wait_window(dsa)
                if dsa.folder is None:
                    self._Q.trigger("exit")
                    return
                res_path = Path(dsa.folder)
            cproc = res_path / "cproc.clog"
            self.tab_trip.load(str(cproc))
        
        # Load the results
        self.set_status(_L["LOADING"])
        self.folder = str(res_path.absolute() / "figures")
        self.title(f'{_L["TITLE"]} - {res_path.name}')
        self.disable_all()

        def load_async(res_path):
            sta = ReadOnlyStatistics(res_path)
            npl = AdvancedPlot()
            npl.load_series(sta)
            return (sta, npl)
        
        self._Q.submit("loaded", load_async, res_path)

        state_path = res_path / SAVED_STATE_FOLDER / TRAFFIC_INST_FILE_NAME

        def load_state_async(state_path:Path):
            try:
                import cloudpickle as pickle
                with gzip.open(state_path, 'rb') as fp:
                    d  = pickle.load(fp)
                assert isinstance(d, dict)
                assert "obj" in d
                inst = d["obj"]
                assert isinstance(inst, TrafficInst)
            except:
                MB.showerror(_L["ERROR"], _L["SAVED_STATE_LOAD_FAILED"])
                inst = None
            return inst

        if state_path.exists():
            self._Q.submit("state_loaded", load_state_async, state_path)

    def plotSelected(self):
        cfg = self._pp.getConfig()
        self.disable_all()
        self.set_status("Plotting all...")
        self._npl.pic_ext = self._pp.cb_ext.get()
        self._npl.plot_title = self._pp.plot_title.get()
        try:
            self._npl.dpi = int(self._pp.entry_dpi.get())
        except:
            MB.showerror(_L["ERROR"], _L["INVALID_DPI"])
            self.enable_all()
            return
        for a in AVAILABLE_ITEMS2:
            if cfg[a]: break
        else:
            MB.showerror(_L["ERROR"], _L["NOTHING_PLOT"])
            self.enable_all()
        
        def work(cfg):
            def todo(plotpage, opt_name):
                getattr(plotpage, "plot_" + opt_name).set("False")

            try:
                for a in AVAILABLE_ITEMS2:
                    if cfg[a]:
                        getattr(self, "_plot_"+a)()
                        if "_" in a: continue
                        self._Q.delegate(todo, self._pp, a)
            except Exception as e:
                traceback.print_exc()
                return e
            return None

        self._Q.submit("plot_done", work, cfg)

    def _plot_scs_accum(self):
        tl,tr = self._pp.getTime()
        self._npl.quick_scs_accum(tl, tr, self._pp.AccumPlotMax, res_path=self._sta.root)
    
    def _plot_fcs_accum(self):
        tl, tr = self._pp.getTime()
        self._npl.quick_fcs_accum(tl, tr, self._pp.AccumPlotMax, res_path=self._sta.root)

    def _plot_fcs(self):
        t = self._pp.fcs_pad.get()
        if t.strip()=="" or t==ITEM_ALL:
            cs = self._sta.FCS_head
        elif t==ITEM_SUM:
            cs = ["<sum>"]
        else:
            cs = [x.strip() for x in t.split(',')]
        for i,c in enumerate(cs,start=1):
            self._Q.delegate(self.set_status, f'({i} of {len(cs)})Plotting FCS graph...')
            self._npl.quick_fcs(
                cs_name=c, res_path=self._sta.root, 
                **self._pp.pars("fcs")
            )

    def _plot_scs(self):
        t = self._pp.scs_pad.get()
        if t.strip()=="" or t==ITEM_ALL:
            cs = self._sta.SCS_head
        elif t==ITEM_SUM:
            cs = ["<sum>"]
        else:
            cs = [x.strip() for x in t.split(',')]
        for i,c in enumerate(cs,start=1):
            self._Q.delegate(self.set_status, f'({i} of {len(cs)})Plotting SCS graph...')
            self._npl.quick_scs(
                cs_name=c, res_path=self._sta.root,
                **self._pp.pars("scs")
            )

    def _plot_ev(self):
        self._npl.tl = int(self._pp.entry_time.get())
        t = self._pp.ev_pad.get()
        evs=None if t.strip()=="" else [x.strip() for x in t.split(',')]
        if evs is None:
            self._Q.trigger("error", 'ID of EV cannot be empty')
            return
        for ev in evs:
            self._npl.quick_ev(ev_name = ev,
                res_path=self._sta.root,
                **self._pp.pars("ev")
            )
    
    def _plot_gen(self):
        t = self._pp.gen_pad.get()
        if t.strip()=="" or t==ITEM_ALL:
            gen = self._sta.gen_head
        elif t==ITEM_ALL_G:
            gen = [x for x in self._sta.gen_head if not x.startswith("V2G")]
        elif t==ITEM_ALL_V2G:
            gen = [x for x in self._sta.gen_head if x.startswith("V2G")]
        else: gen = [x.strip() for x in t.split(',')]
        for i, g in enumerate(gen, start=1):
            self._Q.delegate(self.set_status, f'({i}/{len(gen)})Plotting generators...')
            self._npl.quick_gen(
                gen_name=g,res_path=self._sta.root,
                **self._pp.pars("gen")
            )

    def _plot_bus(self):
        t=self._pp.bus_pad.get()
        if t.strip()=="" or t==ITEM_ALL:
            bus=self._sta.bus_head
        else: bus=[x.strip() for x in t.split(',')]
        for i,g in enumerate(bus,start=1):
            self._Q.delegate(self.set_status, f'({i}/{len(bus)})Plotting buses...')
            self._npl.quick_bus(
                bus_name = g, res_path=self._sta.root,
                **self._pp.pars("bus")
            )
    
    def _plot_gen_total(self):
        tl, tr = self._pp.getTime()
        self._npl.quick_gen_tot(tl,tr,True,True,True,res_path=self._sta.root)
    
    def _plot_bus_total(self):
        tl, tr = self._pp.getTime()
        self._npl.quick_bus_tot(tl,tr,True,True,True,True,res_path=self._sta.root)
    
    def _plot_line(self):
        t=self._pp.line_pad.get()
        if t.strip()=="" or t==ITEM_ALL:
            line=self._sta.line_head
        else: line=[x.strip() for x in t.split(',')]
        for i,g in enumerate(line,start=1):
            self._Q.delegate(self.set_status, f'({i}/{len(line)})Plotting lines...')
            self._npl.quick_line(
                line_name = g, res_path=self._sta.root,
                **self._pp.pars("line")
            )

    def _plot_pvw(self):
        t = self._pp.pvw_pad.get()
        if t.strip() == "" or t == ITEM_ALL:
            pvw = self._sta.pvw_head
        else: pvw = [x.strip() for x in t.split(',')]
        for i, g in enumerate(pvw,start=1):
            self._Q.delegate(self.set_status, f'({i}/{len(pvw)})Plotting PV & Wind...')
            self._npl.quick_pvw(
                pvw_name = g, res_path=self._sta.root,
                **self._pp.pars("pvw")
            )

    def _plot_ess(self):
        t = self._pp.ess_pad.get()
        if t.strip() == "" or t == ITEM_ALL:
            ess = self._sta.ess_head
        else: ess = [x.strip() for x in t.split(',')]
        for i,g in enumerate(ess,start=1):
            self._Q.delegate(self.set_status, f'({i}/{len(ess)})Plotting ESS...')
            self._npl.quick_ess(
                ess_name = g, res_path = self._sta.root,
                **self._pp.pars("ess")
            )

__all__ = ["ViewerBox"]


from feasytools import ArgChecker
from v2simux_gui.welcomebox import WelcomeBox


def main():
    args = ArgChecker()
    to_open = args.pop_str("d", "")
    
    wb = WelcomeBox(to_open)
    msg = wb.show()

    if msg[0] == "close":
        exit(0)
    elif msg[0] == "main" and msg[1] != "":
        from v2simux_gui.mainbox import MainBox
        win = MainBox()
        win.folder = msg[1]
        win._load()
        win.mainloop()
    elif msg[0] == "res" and msg[1] != "":
        from v2simux_gui.viewerbox import ViewerBox
        win = ViewerBox(msg[1])
        win.mainloop()
    elif msg[0] == "conv":
        from v2simux_gui.convertbox import ConvertBox
        win = ConvertBox()
        win.mainloop()
    elif msg[0] == "para":
        from v2simux_gui.parabox import ParaBox
        win = ParaBox()
        win.mainloop()

if __name__ == "__main__":
    main()


from dataclasses import dataclass
from typing import Callable, List, Dict, Tuple, Set
from sklearn.neighbors import KDTree
from .uxsim import Link
import heapq
import math


@dataclass
class Stage:
    nodes: List[str]
    edges: List[str]
    travelTime: float
    length: float

CoordsDict = Dict[str, Tuple[float, float]]  # node ID -> (x, y)
Graph = Dict[str, List[Tuple[str, Link]]]  # node ID -> List of (to_node ID, Link object)


def dijMC(gl: Graph, ctime: int, from_node: str, to_nodes: Set[str], omega: float, 
          to_charge: float, wt: Dict[str, float], p: Dict[str, float], 
          max_length: float = float('inf')) -> Stage:
    """
    Find the BEST route based on score = omega * (time + waiting) + charging_cost.
    Uses time as primary key, length as secondary key.
    """
    # (time, length, node, path, path_edges)
    
    heap = [(0, 0, from_node, [from_node], [])]
    visited = set()
    min_time = {from_node: 0}
    best_score = float('inf')
    best_stage = Stage([], [], float('inf'), float('inf'))

    while heap:
        cur_time, cur_len, cur_node, path, path_edges = heapq.heappop(heap)
        
        if cur_node in visited:
            continue
        visited.add(cur_node)
        
        # 检查是否到达目标节点且满足长度约束
        if cur_node in to_nodes:
            score = omega * (cur_time / 60 + wt[cur_node]) + to_charge * p[cur_node]
            if score < best_score:
                best_score = score
                best_stage = Stage(path.copy(), path_edges.copy(), cur_time, cur_len)
        
        # 探索邻居节点
        for neighbor, link in gl.get(cur_node, []):
            if neighbor in visited:
                continue
                
            next_time = cur_time + link.instant_travel_time(ctime)
            next_len = cur_len + link.length
            
            if next_len > max_length:
                continue
                
            if neighbor not in min_time or next_time < min_time[neighbor]:
                min_time[neighbor] = next_time
                heapq.heappush(heap, (next_time, next_len, neighbor, path + [neighbor], path_edges + [link.name]))
    
    return best_stage

def dijMF(gl: Graph, ctime: int, from_node: str, to_nodes: Set[str]) -> Stage:
    """
    Find the FASTEST route to any node in to_nodes.
    Uses time as primary key, length as secondary key.
    """
    # (time, length, node, path, path_edges)
    heap = [(0, 0, from_node, [from_node], [])]
    visited = set()
    min_time = {from_node: 0}
    best_stage = Stage([], [], float('inf'), float('inf'))
    
    while heap:
        cur_time, cur_len, cur_node, path, path_edges = heapq.heappop(heap)

        if cur_node in visited:
            continue
        visited.add(cur_node)
        
        if cur_node in to_nodes:
            if cur_time < best_stage.travelTime:
                best_stage = Stage(path.copy(), path_edges.copy(), cur_time, cur_len)
        
        for neighbor, link in gl.get(cur_node, []):
            if neighbor in visited:
                continue
                
            next_time = cur_time + link.instant_travel_time(ctime)
            next_len = cur_len + link.length
            
            if neighbor not in min_time or next_time < min_time[neighbor]:
                min_time[neighbor] = next_time
                heapq.heappush(heap, (next_time, next_len, neighbor, path + [neighbor], path_edges + [link.name]))
    
    return best_stage

def dijMS(gl: Graph, ctime: int, from_node: str, to_nodes: Set[str]) -> Stage:
    """
    Find the SHORTEST route to any node in to_nodes.
    Uses length as primary key, time as secondary key.
    """
    # (length, time, node, path, path_edges)
    heap = [(0, 0, from_node, [from_node], [])]
    visited = set()
    min_length = {from_node: 0}
    best_stage = Stage([], [],float('inf'), float('inf'))
    
    while heap:
        cur_len, cur_time, cur_node, path, path_edges = heapq.heappop(heap)
        
        if cur_node in visited:
            continue
        visited.add(cur_node)
        
        if cur_node in to_nodes:
            if cur_len < best_stage.length:
                best_stage = Stage(path.copy(), path_edges.copy(), cur_time, cur_len)

        for neighbor, link in gl.get(cur_node, []):
            if neighbor in visited:
                continue
                
            next_len = cur_len + link.length
            next_time = cur_time + link.instant_travel_time(ctime)
            
            if neighbor not in min_length or next_len < min_length[neighbor]:
                min_length[neighbor] = next_len
                heapq.heappush(heap, (next_len, next_time, neighbor, path + [neighbor], path_edges + [link.name]))
    
    return best_stage

def dijF(gl: Graph, ctime: int, from_node: str, to_node: str) -> Stage:
    """Fastest path between two specific nodes."""
    return dijMF(gl, ctime, from_node, {to_node})

def dijS(gl: Graph, ctime: int, from_node: str, to_node: str) -> Stage:
    """Shortest path between two specific nodes."""
    return dijMS(gl, ctime, from_node, {to_node})

def astarF(gl: Graph, node_coords: CoordsDict, ctime: int, from_node: str, to_node: str) -> Stage:
    """
    A* algorithm for FASTEST path using time as cost.
    Uses f_score (time + heuristic) as primary key, time as secondary key, length as tertiary key.
    """
    to_coord = node_coords[to_node]
    
    def heuristic(node):
        coord = node_coords[node]
        return math.hypot(coord[0] - to_coord[0], coord[1] - to_coord[1])
    
    heap = []
    initial_h = heuristic(from_node)
    # (f_score, time, length, node, path, path_edges)
    heapq.heappush(heap, (initial_h, 0, 0, from_node, [from_node], []))
    
    visited = set()
    g_scores = {from_node: 0}  # g_score = actual travel time
    
    while heap:
        f_score, cur_time, cur_len, cur_node, path, path_edges = heapq.heappop(heap)
        
        if cur_node in visited:
            continue
            
        if cur_node == to_node:
            return Stage(path, path_edges, cur_time, cur_len)
            
        visited.add(cur_node)

        for neighbor, link in gl.get(cur_node, []):
            if neighbor in visited:
                continue
                
            time_cost = link.instant_travel_time(ctime)
            next_time = cur_time + time_cost
            next_len = cur_len + link.length
            
            if neighbor not in g_scores or next_time < g_scores[neighbor]:
                g_scores[neighbor] = next_time
                h_score = heuristic(neighbor)
                f_score = next_time + h_score
                
                heapq.heappush(heap, (f_score, next_time, next_len, neighbor, path + [neighbor], path_edges + [link.name]))
    
    return Stage([], [], float('inf'), float('inf'))

def astarS(gl: Graph, node_coords: CoordsDict, ctime: int, from_node: str, to_node: str) -> Stage:
    """
    A* algorithm for SHORTEST path using length as cost.
    Uses f_score (length + heuristic) as primary key, length as secondary key, time as tertiary key.
    """
    to_coord = node_coords[to_node]
    
    def heuristic(node):
        coord = node_coords[node]
        return math.hypot(coord[0] - to_coord[0], coord[1] - to_coord[1])
    
    heap = []
    initial_h = heuristic(from_node)
    # (f_score, length, time, node, path, path_edges)
    heapq.heappush(heap, (initial_h, 0, 0, from_node, [from_node], []))
    
    visited = set()
    g_scores = {from_node: 0}  # g_score = actual path length
    
    while heap:
        f_score, cur_len, cur_time, cur_node, path, path_edges = heapq.heappop(heap)
        
        if cur_node in visited:
            continue
            
        if cur_node == to_node:
            return Stage(path, path_edges, cur_time, cur_len)
            
        visited.add(cur_node)
        
        for neighbor, link in gl.get(cur_node, []):
            if neighbor in visited:
                continue
                
            length_cost = link.length
            next_len = cur_len + length_cost
            next_time = cur_time + link.instant_travel_time(ctime)
            
            if neighbor not in g_scores or next_len < g_scores[neighbor]:
                g_scores[neighbor] = next_len
                h_score = heuristic(neighbor)
                f_score = next_len + h_score

                heapq.heappush(heap, (f_score, next_len, next_time, neighbor, path + [neighbor], path_edges + [link.name]))
    
    return Stage([], [], float('inf'), float('inf'))

# 构建kDTree的预处理函数
def build_target_kdtree(node_coords: CoordsDict, to_nodes: Set[str]) -> Tuple[KDTree, Dict[int, str]]:
    """
    为目标节点构建kDTree并建立索引到节点ID的映射
    Returns: (kdtree, index_to_node_id)
    """
    target_coords = []
    index_to_node = {}
    
    for idx, node_id in enumerate(to_nodes):
        if node_id in node_coords:
            coord = node_coords[node_id]
            target_coords.append(coord)
            index_to_node[idx] = node_id
    
    kdtree = KDTree(target_coords)
    return kdtree, index_to_node

# 修改后的启发式函数
def create_heuristic_with_kdtree(node_coords: CoordsDict, kdtree: KDTree) -> Callable:
    """
    创建使用kDTree的启发式函数
    """
    def heuristic(node):
        if node not in node_coords or kdtree is None:
            return float('inf')
        
        coord = node_coords[node]
        # 查询最近的一个目标节点
        distance, _ = kdtree.query([coord], k=1)
        return distance[0]
    
    return heuristic

def create_time_heuristic_with_kdtree(node_coords: CoordsDict, kdtree: KDTree, avg_speed: float = 20.0) -> Callable:
    """
    创建使用kDTree的时间启发式函数
    """
    def heuristic(node):
        if node not in node_coords or kdtree is None:
            return float('inf')
        
        coord = node_coords[node]
        # 查询最近的一个目标节点
        distance, _ = kdtree.query([coord], k=1)
        return distance[0] / avg_speed  # 将距离转换为时间估计
    
    return heuristic

# 修改后的A*算法（以astarMF为例）
def astarMF(gl: Graph, node_coords: CoordsDict, ctime: int, from_node: str, to_nodes: Set[str], avg_speed:float) -> Stage:
    """
    A* version of dijMF - Find the FASTEST route to any node in to_nodes using kDTree heuristic.
    """
    # 构建kDTree
    kdtree, index_to_node = build_target_kdtree(node_coords, to_nodes)
    
    # 创建启发式函数
    heuristic = create_time_heuristic_with_kdtree(node_coords, kdtree, avg_speed=avg_speed)
    
    # (estimated_total_time, time, length, node, path, path_edges)
    initial_h = heuristic(from_node)
    heap = [(initial_h, 0, 0, from_node, [from_node], [])]
    
    visited = set()
    g_scores = {from_node: 0}  # g_score = actual travel time
    best_stage = Stage([], [], float('inf'), float('inf'))
    
    while heap:
        est_total_time, cur_time, cur_len, cur_node, path, path_edges = heapq.heappop(heap)
        
        if cur_node in visited:
            continue
            
        if cur_node in to_nodes:
            if cur_time < best_stage.travelTime:
                best_stage = Stage(path.copy(), path_edges.copy(), cur_time, cur_len)
        
        visited.add(cur_node)

        for neighbor, link in gl.get(cur_node, []):
            if neighbor in visited:
                continue
                
            time_cost = link.instant_travel_time(ctime)
            next_time = cur_time + time_cost
            next_len = cur_len + link.length
            
            if neighbor not in g_scores or next_time < g_scores[neighbor]:
                g_scores[neighbor] = next_time
                h_score = heuristic(neighbor)
                est_total = next_time + h_score

                heapq.heappush(heap, (est_total, next_time, next_len, neighbor, path + [neighbor], path_edges + [link.name]))

    return best_stage

def astarMS(gl: Graph, node_coords: CoordsDict, ctime: int, from_node: str, to_nodes: Set[str]) -> Stage:
    """
    A* version of dijMS - Find the SHORTEST route to any node in to_nodes using kDTree heuristic.
    """
    # 构建kDTree
    kdtree, index_to_node = build_target_kdtree(node_coords, to_nodes)
    
    # 创建启发式函数
    heuristic = create_heuristic_with_kdtree(node_coords, kdtree)
    
    # (estimated_total_length, length, time, node, path, path_edges)
    initial_h = heuristic(from_node)
    heap = [(initial_h, 0, 0, from_node, [from_node], [])]
    
    visited = set()
    g_scores = {from_node: 0}  # g_score = actual path length
    best_stage = Stage([], [], float('inf'), float('inf'))
    
    while heap:
        est_total_len, cur_len, cur_time, cur_node, path, path_edges = heapq.heappop(heap)
        
        if cur_node in visited:
            continue
            
        if cur_node in to_nodes:
            if cur_len < best_stage.length:
                best_stage = Stage(path.copy(), path_edges.copy(), cur_time, cur_len)
        
        visited.add(cur_node)

        for neighbor, link in gl.get(cur_node, []):
            if neighbor in visited:
                continue
                
            length_cost = link.length
            next_len = cur_len + length_cost
            next_time = cur_time + link.instant_travel_time(ctime)
            
            if neighbor not in g_scores or next_len < g_scores[neighbor]:
                g_scores[neighbor] = next_len
                h_score = heuristic(neighbor)
                est_total = next_len + h_score
                
                heapq.heappush(heap, (est_total, next_len, next_time, neighbor, path + [neighbor], path_edges + [link.name]))
    
    return best_stage

def astarMC(gl: Graph, node_coords: CoordsDict, ctime: int, from_node: str, to_nodes: Set[str], 
           omega: float, to_charge: float, wt: Dict[str, float], p: Dict[str, float], 
           max_length: float = float('inf'), avg_speed: float = 20.0) -> Stage:
    """
    A* version of dijMC - Find the BEST route using kDTree heuristic.
    """
    # 构建kDTree
    kdtree, index_to_node = build_target_kdtree(node_coords, to_nodes)
    
    # 创建启发式函数（使用距离启发式）
    distance_heuristic = create_heuristic_with_kdtree(node_coords, kdtree)
    
    # 将距离启发式转换为score启发式（近似）
    def score_heuristic(node):
        dist = distance_heuristic(node)
        # 将距离转换为近似的score增量（这是一个保守估计）
        return omega * (dist / avg_speed / 60.0)  # 假设平均速度20.0m/s，转换为分钟
    
    # (estimated_total_score, actual_score, time, length, node, path, path_edges)
    initial_h = score_heuristic(from_node)
    initial_score = omega * (0 / 60.0 + wt.get(from_node, 0)) + to_charge * p.get(from_node, 0)
    heap = [(initial_h + initial_score, initial_score, 0, 0, from_node, [from_node], [])]
    
    visited = set()
    best_score = {from_node: initial_score}
    best_stage = Stage([], [], float('inf'), float('inf'))
    min_actual_score = float('inf')
    
    while heap:
        est_total_score, cur_score, cur_time, cur_len, cur_node, path, path_edges = heapq.heappop(heap)

        if cur_score > min_actual_score:
            continue
            
        if cur_node in visited:
            continue
        visited.add(cur_node)
        
        if cur_node in to_nodes and cur_len <= max_length:
            if cur_score < min_actual_score:
                min_actual_score = cur_score
                best_stage = Stage(path.copy(), path_edges.copy(), cur_time, cur_len)

        for neighbor, link in gl.get(cur_node, []):
            if neighbor in visited:
                continue
                
            next_time = cur_time + link.instant_travel_time(ctime)
            next_len = cur_len + link.length
            
            if next_len > max_length:
                continue
            
            neighbor_score = omega * (next_time / 60.0 + wt.get(neighbor, 0)) + to_charge * p.get(neighbor, 0)
            
            if neighbor not in best_score or neighbor_score < best_score[neighbor]:
                best_score[neighbor] = neighbor_score
                h_score = score_heuristic(neighbor)
                est_total = neighbor_score + h_score
                
                heapq.heappush(heap, (est_total, neighbor_score, next_time, next_len, 
                                    neighbor, path + [neighbor], path_edges + [link.name]))

    return best_stage


import operator
from itertools import chain
from feasytools import LangLib
from ..traffic import VehStatus
from .base import *


FILE_EV = "ev"
EV_ATTRIB = ["soc", "status", "cost", "earn", "x", "y"]
_L = LangLib(["en", "zh_CN"])
_L.SetLangLib("en", EV = "EV")
_L.SetLangLib("zh_CN", EV = "电动车")


class StaEV(StaBase):
    def __init__(self, path:str, tinst:TrafficInst, plugins:Dict[str,PluginBase]):
        super().__init__(FILE_EV, path, cross_list(tinst.vehicles.keys(), EV_ATTRIB), tinst, plugins)

    @staticmethod
    def GetLocalizedName() -> str:
        return _L("EV")
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return []
    
    def GetData(self, inst:TrafficInst, plugins:Dict[str,PluginBase]) -> Iterable[Any]:
        vehs = inst.vehicles.values()
        soc = (veh.SOC for veh in vehs)
        status = map(operator.attrgetter("_sta"), vehs)
        cost = map(operator.attrgetter("_cost"), vehs)
        earn = map(operator.attrgetter("_earn"), vehs)
        x = []; y = []
        for veh in inst.vehicles.values():
            if veh.status == VehStatus.Driving:
                pos = inst.W.get_vehicle(veh._id).get_xy_coords()
            else:
                pos = (0, 0)
            x.append(pos[0])
            y.append(pos[1])
        return chain(soc, status, cost, earn, x, y)


from pathlib import Path
from feasytools import ArgChecker
from v2simux_gui.viewerbox import ViewerBox
    
if __name__ == "__main__":
    args = ArgChecker()
    dir = args.pop_str("d", "")
    if dir != "" and not Path(dir).is_dir():
        raise FileNotFoundError(f"{dir} is not a directory.")
    
    win = ViewerBox(dir)
    win.mainloop()


from .base import (
    PluginBase, PluginStatus, PluginConfigItem, IGridPlugin, EditMode,
    Getter, Setter, Validator, ConfigDict, PIResult, PIExec, PINoRet,
    ConfigItem, PluginConfigItem, ConfigItemDict
)
from .pdn import PluginPDN
from .v2g import PluginV2G
from .ocur import PluginOvercurrent
from .pool import PluginPool, PluginError, PluginMan


from typing import Dict
from feasytools import RangeList
from ..locale.lang import Lang
from .utils import ReadXML
from .params import *
from .ev import EV, Trip

class FloatDictWrapper:
    def __init__(self, d: Dict[str, str]):
        self.__d = d
    
    def get(self, key: str, default: float) -> float:
        return float(self.__d.get(key, default))
    
class EVDict(Dict[str, EV]):
    def __init__(self, file_path=None):
        super().__init__()
        if file_path is None:
            return
        rt = ReadXML(file_path).getroot()
        if rt is None:
            raise RuntimeError(Lang.EV_LOAD_ERROR.format(file_path))
        for veh in rt:
            trips: list[Trip] = []            
            for trip in veh:
                new_trip = Trip(
                    trip.attrib.pop("id"),
                    int(float(trip.attrib.pop("depart"))),
                    trip.attrib.pop("fromNode"),
                    trip.attrib.pop("toNode"),
                )
                if len(trips) > 0:
                    if new_trip.depart_time <= trips[-1].depart_time:
                        raise ValueError(Lang.BAD_TRIP_DEPART_TIME.format(
                            new_trip.depart_time, trips[-1].depart_time, veh.attrib['id'], new_trip.id))
                    if new_trip.from_node != trips[-1].to_node:
                        raise ValueError(Lang.BAD_TRIP_OD.format(
                            new_trip.from_node, trips[-1].to_node, veh.attrib['id'], new_trip.id))
                trips.append(new_trip)
            attr = FloatDictWrapper(veh.attrib)
            elem_sctime = veh.find("sctime")
            elem_v2gtime = veh.find("v2gtime")
            self.add(EV(
                veh.attrib["id"],
                trips,
                attr.get("eta_c", DEFAULT_ETA_CHARGE),
                attr.get("eta_d", DEFAULT_ETA_DISCHARGE),
                attr.get("bcap", DEFAULT_FULL_BATTERY),
                attr.get("soc", DEFAULT_INIT_SOC),
                attr.get("c", DEFAULT_CONSUMPTION),
                attr.get("rf", DEFAULT_FAST_CHARGE_RATE),
                attr.get("rs", DEFAULT_SLOW_CHARGE_RATE),
                attr.get("rv", DEFAULT_MAX_V2G_RATE),
                attr.get("omega", DEFAULT_OMEGA),
                attr.get("kr", DEFAULT_KREL),
                attr.get("kf", DEFAULT_FAST_CHARGE_THRESHOLD),
                attr.get("ks", DEFAULT_SLOW_CHARGE_THRESHOLD),
                attr.get("kv", DEFAULT_KV2G),
                veh.attrib.get("rmod", DEFAULT_RMOD),
                RangeList(elem_sctime),
                attr.get("max_sc_cost", DEFAULT_MAX_SC_COST),
                RangeList(elem_v2gtime),
                attr.get("min_v2g_earn", DEFAULT_MIN_V2G_EARN),
                veh.attrib.get("cache_route", "false").lower() == "true",
            ))

    def add(self, ev: EV):
        """Add a vehicle"""
        super().__setitem__(ev.ID, ev)

    def pop(self, veh_id: str) -> EV:
        """
        Remove a vehicle by ID, return the removed value
        """
        return super().pop(veh_id)

__all__ = ["EVDict"]


from v2simux_gui.common import *

from feasytools import PDFunc, CreatePDFunc
from .scrtv import ScrollableTreeView


_L = LangLib.Load(__file__)


def _removeprefix(s: str, prefix: str) -> str:
    if s.startswith(prefix):
        return s[len(prefix):]
    return s

class PDFuncEditor(Toplevel):
    def reset_tree(self, pdfunc:PDFunc):
        self.tree.clear()
        for l,r in pdfunc.__dict__.items():
            self.tree.insert("", "end", values=(l, r))
    
    def __init__(self, var: StringVar):
        super().__init__()
        self.title(_L["PDFUNC_EDITOR"])
        pdfunc = eval(var.get())
        assert isinstance(pdfunc, PDFunc)
        self.model = StringVar(self, _removeprefix(pdfunc.__class__.__name__, "PD"))
        self.fr0 = Frame(self)
        self.mlabel = Label(self.fr0, text=_L["PDMODEL"])
        self.mlabel.grid(row=0,column=0,padx=3,pady=3,sticky="w")
        self.cb = Combobox(self.fr0, textvariable=self.model,
            values=["Normal", "Uniform", "Triangular", 
            "Exponential", "Gamma", "Weibull", 
            "Beta", "LogNormal", "LogLogistic"])
        self.cb.bind('<<ComboboxSelected>>', lambda x: self.reset_tree(CreatePDFunc(self.model.get())))
        self.cb.grid(row=0,column=1,padx=3,pady=3,sticky="w")
        self.fr0.pack(fill="x", expand=True)
        self.tree = ScrollableTreeView(self, allowSave=False)
        self.tree['show'] = 'headings'
        self.tree["columns"] = ("t", "d")
        self.tree.column("t", width=120, stretch=NO)
        self.tree.column("d", width=120, stretch=YES)
        self.tree.heading("t", text=_L["PROPERTY"])
        self.tree.heading("d", text=_L["VALUE"])
        self.tree.pack(fill="both", expand=True)
        self.reset_tree(pdfunc)
        self.tree.setColEditMode("d", ConfigItem(name="d", editor=EditMode.ENTRY, desc="Value"))
        self.fr = Frame(self)
        self.fr.pack(fill="x", expand=False)
        self.btn_save = Button(self.fr, text=_L["SAVE_AND_CLOSE"], command=self.save)
        self.btn_save.grid(row=0,column=4,padx=3,pady=3,sticky="e")
        self.var = var

    def getAllData(self) -> PDFunc:
        res:Dict[str, float] = {}
        for i in self.tree.get_children():
            x = self.tree.tree.item(i, "values")
            res[x[0]] = float(x[1])
        return CreatePDFunc(self.model.get(), **res)
    
    def save(self):
        d = self.getAllData()
        self.var.set(repr(d))
        self.destroy()

__all__ = ["PDFuncEditor"]


from v2simux_gui.com_no_vx import *
from v2simux_gui.langhelper import *
import os


_ = LangLib.Load(__file__)
RECENT_PROJECTS_FILE = Path.home() / ".v2simux" / "recent_projects.txt"


class WelcomeBox(Tk):
    def __init__(self, to_open:str=""):
        super().__init__()
        self.wm_attributes('-topmost',1)
        self.title(_("WELCOME"))
        self.update_idletasks()
        width = 600
        height = 350
        x = (self.winfo_screenwidth() // 2) - (width // 2)
        y = (self.winfo_screenheight() // 2) - (height // 2)
        self.geometry(f'{width}x{height}+{x}+{y}')

        self.menu = Menu(self)
        self.config(menu=self.menu)
        add_lang_menu(self.menu)
        menuTools = Menu(self.menu, tearoff=0)
        menuTools.add_command(label=_("PARA"), command=self._goto_para)
        menuTools.add_separator()
        menuTools.add_command(label=_("HELP"), command=self.open_help)
        menuTools.add_command(label=_("ABOUT"), command=self.show_about)
        self.menu.add_cascade(label=_("TOOLS"), menu=menuTools)

        # Header
        header_frame = Frame(self)
        header_frame.pack(side="top", fill="x", pady=(20, 10))

        # Image
        try:
            self.logo_img = PhotoImage(file=str(Path(__file__).parent / "v2sim.png"))
            logo_label = Label(header_frame, image=self.logo_img)
            logo_label.pack(side="left", padx=(20, 10))
        except Exception:
            logo_label = Label(header_frame, text="[Logo]")
            logo_label.pack(side="left", padx=(20, 10))

        title_label = Label(header_frame, text=_("WELCOME"), font=("Arial", 20, "bold"))
        title_label.pack(side="left", padx=10)

        # Middle section
        middle_frame = Frame(self)
        middle_frame.pack(expand=True, fill="both", pady=10, padx=20)
        
        recent_label = Label(middle_frame, text=_("RECENT_PROJ"))
        recent_label.grid(row=0, column=0, sticky="w", pady=(0, 5))

        self.recent_var = StringVar()
        self.proj_dir = Entry(middle_frame, textvariable=self.recent_var, state="readonly", width=40)
        self.proj_dir.grid(row=0, column=1, columnspan=2, sticky="ew", padx=(5, 0), pady=(0, 5))

        self.project_list = Listbox(middle_frame, height=6)
        self.project_list.grid(row=1, column=0, columnspan=3, sticky="nsew", pady=(0, 5))

        def on_project_select(event):
            selection = self.project_list.curselection()
            if selection:
                selected_project = self.project_list.get(selection[0])
                self.recent_var.set(selected_project)

        self.project_list.bind("<<ListboxSelect>>", on_project_select)
        
        middle_frame.rowconfigure(1, weight=1)
        middle_frame.columnconfigure(1, weight=1)

        self.load_recent_projects()
        if os.path.isdir(to_open):
            self.project_list.insert(0, str(Path(to_open).absolute()))
            self.project_list.selection_set(0)
            on_project_select(None)
            self.save_recent_project()
        
        def clear_list(e):
            if MB.askyesno(_("CONFIRM"), _("CONFIRM_TEXT")):
                self.project_list.delete(0, END)
                self.recent_var.set("")
                if RECENT_PROJECTS_FILE.exists():
                    RECENT_PROJECTS_FILE.unlink()
                self.save_recent_project()

        def select_project(e):
            folder = filedialog.askdirectory(title=_("SELECT_PROJ_FOLDER"))
            if folder:
                self.project_list.insert(0, folder)
                while self.project_list.size() > 10:
                    self.project_list.delete(10)
                self.project_list.selection_clear(0, END)
                self.project_list.select_set(0)
                self.project_list.see(0)
                self.recent_var.set(folder)

                self.save_recent_project()

        # Links
        self.links_panel = Frame(middle_frame)
        self.links_panel.grid(row=2, column=0, columnspan=2, sticky="w")

        self.select_linklbl = Label(
            self.links_panel,
            text=_("ADD_PROJ"),
            foreground="blue",
            cursor="hand2",
            font=("Arial", 10, "underline")
        )
        self.select_linklbl.grid(row=0, column=0, sticky="w", pady=(5, 0))
        self.select_linklbl.bind("<Button-1>", select_project)
        
        self.convert_linklbl = Label(
            self.links_panel,
            text=_("CONV_CASE"),
            foreground="blue",
            cursor="hand2",
            font=("Arial", 10, "underline")
        )
        self.convert_linklbl.grid(row=0, column=1, sticky="w", pady=(5, 0), padx=(10, 0))
        self.convert_linklbl.bind("<Button-1>", self._convert_case)

        self.clear_linklbl = Label(
            self.links_panel,
            text=_("CLEAR_LIST"),
            foreground="blue",
            cursor="hand2",
            font=("Arial", 10, "underline")
        )
        self.clear_linklbl.grid(row=0, column=2, sticky="w", pady=(5, 0), padx=(10, 0))
        self.clear_linklbl.bind("<Button-1>", clear_list)

        # Buttons
        self.btn_panel = Frame(middle_frame)
        self.btn_panel.grid(row=2, column=2, sticky="ew")
        self.view_res_btn = Button(self.btn_panel, text=_("VIEW_RESULTS"), command=self._view_results)
        self.view_res_btn.grid(row=0, column=0, sticky="e", pady=(5, 0), padx=(5, 0))
        self.open_btn = Button(self.btn_panel, text=_("OPEN"), command=self._close)
        self.open_btn.grid(row=0, column=1, sticky="e", pady=(5, 0))

        self.__load_error = False
        self.result = ("close", None)
    
    def open_help(self):
        import webbrowser
        webbrowser.open("https://hesl-seu.github.io/v2sim-wiki/")
        
    def show_about(self):
        import v2simux
        MB.showinfo(_("ABOUT"), _("ABOUT_TEXT").format(v2simux.__version__))

    def _destory(self):
        self.withdraw()
        self.quit()
        self.destroy()
    
    def _check_selected(self):
        if self.recent_var.get() == "":
            MB.showwarning(_("WARNING"), _("PLS_SELECT_PROJ"))
            return False
        return True

    def _goto_para(self):
        self.result = ("para", None)
        self._destory()

    def _close(self):
        if self.__load_error:
            self._destory()
            return
        if not self._check_selected(): return
        self.result = ("main", self.recent_var.get())
        self._destory()
    
    def _view_results(self):
        if not self._check_selected(): return
        self.result = ("res", self.recent_var.get())
        self._destory()
    
    def _convert_case(self, event):
        self.result = ("conv", None)
        self._destory()
    
    def show(self):
        self.update()
        self.deiconify()
        self.mainloop()
        return self.result
    
    def load_recent_projects(self):
        self.project_list.delete(0, END)
        if not RECENT_PROJECTS_FILE.exists():
            return
        with open(RECENT_PROJECTS_FILE, 'r') as f:
            path_list = []
            for line in f:
                project_path = line.strip()
                if project_path and os.path.isdir(project_path):
                    abs_path = str(Path(project_path).absolute().resolve().as_posix())
                    if abs_path not in path_list:
                        path_list.append(abs_path)
                        self.project_list.insert(END, abs_path)

    def save_recent_project(self):
        with open(RECENT_PROJECTS_FILE, 'w') as f:
            for project in self.project_list.get(0, END):
                f.write(f"{project}\n")


# Load Path
from pathlib import Path

# Load Event Queue
from .evtq import EventQueue

# Load Language Library
from feasytools import LangLib, LangConfig
LangConfig.SetAppName("v2simux")
 
# High DPI awareness for Windows
import platform
if platform.system() == "Windows":
    import ctypes
    ctypes.windll.shcore.SetProcessDpiAwareness(1)

# Load Tk controls
from tkinter import (
    Toplevel, messagebox as MB, BooleanVar, StringVar, IntVar, Canvas, Event, Tk, Menu, filedialog, Text, Listbox, PhotoImage, Widget,
    NO, YES, NORMAL, DISABLED, END, BOTH, X, Y, LEFT, RIGHT, TOP, BOTTOM, W, E
)
from tkinter.ttk import Treeview, Button, LabelFrame, Checkbutton, Combobox, Frame, Label, Entry, Spinbox, Scrollbar, Radiobutton, Notebook, OptionMenu

# Load Type Hints
from typing import Dict, List, Set, Tuple, Any, Union, Optional, Iterable, Callable, Literal

# Set exports
__all__ = [
    "LangLib", "LangConfig", "EventQueue", "Path", "Dict", "List", "Set", "Tuple", "Any", "Union", "Optional", "Iterable", "Callable", "Literal",
    "Toplevel", "MB", "BooleanVar", "StringVar", "IntVar", "Canvas", "Event", "Tk", "Menu", "filedialog", "Text", "Listbox", "PhotoImage", "Widget",
    "NO", "YES", "NORMAL", "DISABLED", "END", "X", "Y", "BOTH", "LEFT", "RIGHT", "TOP", "BOTTOM", "W", "E",
    "Treeview", "Button", "LabelFrame", "Checkbutton", "Combobox", "Frame", "Label", "Entry", "Spinbox", "Scrollbar", "Radiobutton", "Notebook", "OptionMenu",
]


import random, string, gzip, sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Set, Dict, List, Tuple, Union
from xml.etree import ElementTree as ET
from ..locale import Lang

SAVED_STATE_FOLDER = "saved_state"

IntPairList = List[Tuple[int, int]]
PriceList = Tuple[List[int], List[float]]
_letters = string.ascii_letters + string.digits


def random_string(length: int):
    return "".join(random.choice(_letters) for _ in range(length))


def ReadXML(file: str, compressed:Optional[bool]=None) -> ET.ElementTree:
    '''
    Read XML file, support compressed GZ file
        file: file path
        compressed: whether the file is compressed. If None, the function will detect it, but only .xml and .xml.gz are supported.
    '''
    filel = file.lower()
    if filel.endswith(".xml.gz") or compressed == True:
        with gzip.open(file, "rt", encoding="utf8") as f:
            return ET.ElementTree(file=f)
    elif filel.endswith(".xml") or compressed == False:
        return ET.ElementTree(file=file)
    else:
        raise RuntimeError(Lang.ERROR_FILE_TYPE_NOT_SUPPORTED.format(file))

def LoadFCS(filename: str) -> Set[str]:
    '''Load FCS file and return a set of node names'''
    fcs_root = ReadXML(filename).getroot()
    if fcs_root is None:
        raise RuntimeError(Lang.ERROR_FILE_TYPE_NOT_SUPPORTED.format(filename))
    fcs_nodes = set()
    for fcs in fcs_root:
        if fcs.tag == "fcs":
            fcs_nodes.add(fcs.attrib["node"])
    return fcs_nodes

def LoadSCS(filename: str) -> Set[str]:
    '''Load SCS file and return a set of node names'''
    scs_root = ReadXML(filename).getroot()
    if scs_root is None:
        raise RuntimeError(Lang.ERROR_FILE_TYPE_NOT_SUPPORTED.format(filename))
    scs_nodes = set()
    for scs in scs_root:
        if scs.tag == "scs":
            scs_nodes.add(scs.attrib["node"])
    return scs_nodes

def CheckFile(file: str):
    p = Path(file)
    if p.exists():
        i = 1
        while True:
            p = Path(file + f".bak{i}")
            i += 1
            if not p.exists():
                break
        Path(file).rename(str(p))

def ClearBakFiles(dir: str):
    for x in Path(dir).iterdir():
        if not x.is_file():
            continue
        if x.suffix == ".bak":
            x.unlink()

@dataclass
class FileDetectResult:
    name: str
    fcs: Optional[str] = None
    scs: Optional[str] = None
    grid: Optional[str] = None
    net: Optional[str] = None
    veh: Optional[str] = None
    plg: Optional[str] = None
    cfg: Optional[str] = None
    taz: Optional[str] = None
    py: Optional[str] = None
    node_type: Optional[str] = None
    osm: Optional[str] = None
    poly: Optional[str] = None
    cscsv: Optional[str] = None
    pref: Optional[str] = None
    poi: Optional[str] = None
    saved_state: Optional[str] = None
    last_result_state: Optional[str] = None
    
    def __getitem__(self, key: str):
        return getattr(self, key)
    
    def has(self, key: str) -> bool:
        return hasattr(self, key)
    
    def get(self, key: str) -> Optional[str]:
        return getattr(self, key, None)
    
    def __contains__(self, key: str) -> bool:
        return hasattr(self, key) and getattr(self, key) != None

@dataclass
class AddtionalTypes:
    Poly: bool
    Poi: bool
    Taz: bool

def CheckAddtionalType(file: str) -> AddtionalTypes:
    root = ReadXML(file, compressed=False).getroot()
    if root is None:
        raise RuntimeError(Lang.ERROR_FILE_TYPE_NOT_SUPPORTED.format(file))
    poly = root.find("poly") is not None
    poi = root.find("poi") is not None
    taz = root.find("taz") is not None
    return AddtionalTypes(Poly=poly, Poi=poi, Taz=taz)

def DetectFiles(dir: Union[str, Path]) -> FileDetectResult:
    """
    Detect simulation-realted files (SUMO config, SCS, FCS, power grid, etc.) in the given directory.
    Args:
        dir (str): Directory path
    Returns:
        FileDetectResult: A dictionary containing the detected files
    """
    p = Path(dir) if isinstance(dir, str) else dir
    ret: Dict[str, str] = {"name": p.name}
    def add(name: str, filename: str):
        if name in ret: raise FileExistsError(Lang.ERROR_CONFIG_DIR_FILE_DUPLICATE.format(name,ret[name],filename))
        ret[name] = filename
    addtional: Set[str] = set()
    for x in p.iterdir():
        if not x.is_file():
            continue
        filename = str(x)
        filenamel = filename.lower()
        if filenamel.endswith(".fcs.xml") or filenamel.endswith(".fcs.xml.gz"):
            add("fcs", filename)
        elif filenamel.endswith(".scs.xml") or filenamel.endswith(".scs.xml.gz"):
            add("scs", filename)
        elif filenamel.endswith(".grid.zip") or filenamel.endswith(".grid.xml"):
            add("grid", filename)
        elif filenamel.endswith(".net.xml") or filenamel.endswith(".net.xml.gz"):
            add("net", filename)
        elif filenamel.endswith(".veh.xml") or filenamel.endswith(".veh.xml.gz"):
            add("veh", filename)
        elif filenamel.endswith(".plg.xml") or filenamel.endswith(".plg.xml.gz"):
            add("plg", filename)
        elif filenamel.endswith(".py"):
            add("py",filename)
        elif filenamel.endswith("node_type.txt"):
            add("node_type", filename)
        elif filenamel.endswith(".osm.xml") or filenamel.endswith(".osm.xml.gz"):
            add("osm", filename)
        elif filenamel.endswith("cs.csv"):
            add("cscsv", filename)
        elif filenamel.endswith(".v2simcfg"):
            add("pref", filename)
        elif (filenamel.endswith(".add.xml") or filenamel.endswith(".add.xml.gz") or
            filenamel.endswith(".poly.xml") or filenamel.endswith(".poly.xml.gz") or
            filenamel.endswith(".taz.xml") or filenamel.endswith(".taz.xml.gz")):
            addtional.add(Path(filename).absolute().as_posix())

    for a in addtional:
        aret = CheckAddtionalType(a)
        if aret.Poly:
            add("poly", a)
        if aret.Poi:
            add("poi", a)
        if aret.Taz:
            add("taz", a)
    
    if (p / SAVED_STATE_FOLDER).exists():
        add("saved_state", (p / SAVED_STATE_FOLDER).as_posix())
    
    if (p / "results" / SAVED_STATE_FOLDER).exists():
        add("last_result_state", (p / "results" / SAVED_STATE_FOLDER).as_posix())

    return FileDetectResult(**ret)

@dataclass
class V2SimConfig:
    start_time: int = 0
    end_time: int = 172800
    break_time: int = 172800
    traffic_step: int = 10
    seed: int = 0
    routing_method:str = "astar"
    load_state: int = 0
    save_state_on_abort: bool = False
    save_state_on_finish: bool = False
    copy_state: bool = False
    visualize: bool = False
    disable_parallel: bool = False
    show_uxsim_info: bool = False
    stats: Optional[List[str]] = None

    @staticmethod
    def load(file:str) -> 'V2SimConfig':
        """
        Load V2Sim configuration from a file.
        Args:
            file (str): Path to the configuration file
        Returns:
            V2SimConfig: A V2SimConfig object with the loaded configuration
        """
        import json
        with open(file, "r") as f:
            data = json.load(f)
        # Remove deprecated fields if present
        if "force_caching" in data:
            del data["force_caching"]
        if "force_nogil" in data:
            del data["force_nogil"]
        if "inital_state_dir" in data: # remove typoed field
            del data["inital_state_dir"]
        return V2SimConfig(**data)
    
    def save(self, file:str):
        """
        Save V2Sim configuration to a file.
        Args:
            file (str): Path to the configuration file
        """
        import json
        with open(file, "w") as f:
            json.dump(self.__dict__, f, indent=4)


def PyVersion() -> Tuple[int, int, int, bool]:
    ver_info = sys.version_info
    has_gil = sys._is_gil_enabled() if hasattr(sys, "_is_gil_enabled") else True # type: ignore
    return (ver_info.major, ver_info.minor, ver_info.micro, has_gil)

def CheckPyVersion(ver:Tuple[int, int, int, bool]) -> bool:
    cur_ver = PyVersion()
    # Allow micro version difference
    return ver[0] == cur_ver[0] and ver[1] == cur_ver[1] and ver[3] == cur_ver[3]

__all__ = [
    "IntPairList", "PriceList", "FileDetectResult", "V2SimConfig", "PyVersion", "CheckPyVersion",
    "DetectFiles", "CheckFile", "ClearBakFiles", "ReadXML", "LoadFCS", "LoadSCS", "SAVED_STATE_FOLDER"
]


from v2simux_gui.common import *

from v2simux import StaPool, StaBase


class LogItemPad(LabelFrame):
    def __init__(self, master, title:str, stapool:StaPool, **kwargs):
        super().__init__(master, text=title, **kwargs)
        self._bvs:Dict[str, BooleanVar] = {}
        self._cbs:Dict[str, Checkbutton] = {}
        self.__stapool = stapool
        for id, val in zip(stapool.GetAllLogItem(), stapool.GetAllLogItemLocalizedName()):
            bv = BooleanVar(self, True)
            self._bvs[id] = bv
            cb = Checkbutton(self, text=val, variable=bv)
            cb.pack(anchor='w', side='left')
            self._cbs[id] = cb
            
    def __getitem__(self, key:str):
        return self._bvs[key].get()
    
    def __setitem__(self, key:str, val:bool):
        self._bvs[key].set(val)
    
    def enable(self, key:str):
        return self._cbs[key].configure(state="enabled")

    def disable(self, key:str):
        return self._cbs[key].configure(state="disabled")
    
    def setEnabled(self, key:str, v:bool):
        if v:
            return self._cbs[key].configure(state="enabled")
        else:
            return self._cbs[key].configure(state="disabled")
    
    def getSelected(self):
        return [k for k, v in self._bvs.items() if v.get()]
    
    def check_by_enabled_plugins(self, enabled_plugins:Iterable[str]):
        p = set(enabled_plugins)
        for k in self._bvs.keys():
            sta_type = self.__stapool.Get(k)
            assert issubclass(sta_type, StaBase)
            deps = sta_type.GetPluginDependency()
            for d in deps:
                if d not in p:
                    self.disable(k)
                    self._bvs[k].set(False)
                    break
            else:
                self.enable(k)
    
    def __contains__(self, key:str):
        return key in self._bvs

__all__ = ["LogItemPad"]


# 关闭找不到路径的错误提示
# Suppress route not found error
SUPPRESS_ROUTE_NOT_FOUND = True

# 默认快充站的充电桩数量, 如果*.fcs.xml中定义了快充站参数，则不会使用此项
# Default number of charging piles in the fast charging station, this item will not be used if the fast charging station parameters are defined in *.fcs.xml
DEFAULT_CS_SLOTS = 20

# 默认路边慢充的充电桩数量, 如果*.scs.xml中定义了路边慢充参数，则不会使用此项
# Default number of charging piles on the roadside slow charging, this item will not be used if the roadside slow charging parameters are defined in *.scs.xml
DEFAULT_PK_SLOTS = 20

# 默认购电价格, 如果*.fcs.xml或*.scs.xml中定义了购电价格，则对应的快充站或路边慢充不会使用此项
# Default purchase price, if the purchase price is defined in *.fcs.xml or *.scs.xml, the corresponding fast charging station or roadside slow charging will not use this item
DEFAULT_BUY_PRICE = ([0], [1.5])

# 默认售电价格, 如果*.fcs.xml或*.scs.xml中定义了购电价格，则对应的快充站或路边慢充不会使用此项
# Default selling price, if the purchase price is defined in *.fcs.xml or *.scs.xml, the corresponding fast charging station or roadside slow charging will not use this item
DEFAULT_SELL_PRICE = ([0], [1.0])

# 默认快充速率kW, 1 kW=1/3600 kWh/s, 如果*.veh.xml或自动生成机构中定义了每辆EV的快充速率，则不会使用此项
# Default fast charge rate kW. This item will not be used if the fast charge rate of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_FAST_CHARGE_RATE = 120 

# 默认慢充速率kW, 如果*.veh.xml或自动生成机构中定义了每辆EV的慢充速率，则不会使用此项
# Default slow charge rate kW. This item will not be used if the slow charge rate of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_SLOW_CHARGE_RATE = 7

# 默认电池容量kWh, 如果*.veh.xml或自动生成机构中定义了每辆EV的电池容量，则不会使用此项
# Default battery capacity kWh. This item will not be used if the battery capacity of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_FULL_BATTERY = 50

# 默认初始电池SoC, 如果*.veh.xml或自动生成机构中定义了每辆EV的初始电池SoC，则不会使用此项
# Default initial battery SoC, this item will not be used if the initial battery SoC of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_INIT_SOC = 0.9

# 默认耗电速率Wh/m, 如果*.veh.xml或自动生成机构中定义了每辆EV的耗电速率，则不会使用此项
# Default power consumption rate Wh/m. This item will not be used if the power consumption rate of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_CONSUMPTION = 320 / DEFAULT_FULL_BATTERY

# 默认用户omega参数, 如果*.veh.xml或自动生成机构中定义了每辆EV的omega，则不会使用此项
# Default user omega parameter, this item will not be used if the omega of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_OMEGA = 75

# 默认用户K_rel参数, 如果*.veh.xml或自动生成机构中定义了每辆EV的Krel，则不会使用此项
# Default user K_rel parameter, this item will not be used if the Krel of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_KREL = 1.25

# 默认用户K_V2G参数, 如果*.veh.xml或自动生成机构中定义了每辆EV的Kv2g，则不会使用此项
# Default user K_V2G parameter, this item will not be used if the Kv2g of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_KV2G = 0.8

# 默认最大V2G返送功率kW, 如果*.veh.xml或自动生成机构中定义了每辆EV的max_v2g_rate，则不会使用此项
# Default maximum V2G return power by kW. This item will not be used if the max_v2g_rate of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_MAX_V2G_RATE = 20

# 指示根据电量是否满足行程长度来充电
# Indicate whether to charge according to whether the power is sufficient for the trip length
ENABLE_DIST_BASED_CHARGING_DECISION = False

# 默认快充阈值: 指定车辆出发时EV电量低于多少比例会去快充, 有效值0~1. 只在ENABLE_DIST_BASED_CHARGING_DECISION = False时生效
# 如果*.veh.xml或自动生成机构中定义了每辆EV的Kfc，则不会使用此项
# Default fast charge threshold: specifies the proportion of EV power when the vehicle departs that will go to fast charge, valid values 0~1. Only valid when ENABLE_DIST_BASED_CHARGING_DECISION = False
# This item will not be used if the Kfc of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_FAST_CHARGE_THRESHOLD = 0.2

# 默认快充阈值: 指定车辆出发时EV电量低于多少比例会去慢充, 有效值0~1.
# 如果*.veh.xml或自动生成机构中定义了每辆EV的Ksc，则不会使用此项
# Default slow charge threshold: specifies the proportion of EV power when the vehicle departs that will go to slow charge, valid values 0~1
# This item will not be used if the Ksc of each EV is defined in *.veh.xml or automatically generated organization
DEFAULT_SLOW_CHARGE_THRESHOLD = 0.5

# 指示在快充站是否根据行程长度确定充电量, True表示充电量刚好够行驶到终点(会稍微多充一点, 由Krel控制), False表示总是充满
# Indicate whether the charging quantity is determined based on the trip length at the fast charging station, True means that the charging quantity is just enough to drive to the end point (a little more charge, controlled by Krel), False means always full
ENABLE_DIST_BASED_CHARGING_QUANTITY = False

# 指示是否允许CSList中同时存在快充站和慢充站
# Indicate whether fast charging stations and slow charging stations are allowed to exist simultaneously in CSList
ALLOW_MIXED_CSTYPE_IN_CSLIST = False

# 默认充电效率
# Default charging efficiency
DEFAULT_ETA_CHARGE = 0.9

# 默认放电效率
# Default discharge efficiency
DEFAULT_ETA_DISCHARGE = 0.9

# 默认充电站所属母线名称
# Default bus name of charging station
DEFAULT_BUS_NAME = "BusAny"

# 默认电动车充电功率修正函数
# Default electric vehicle charging power correction function
DEFAULT_RMOD = "Linear"

# Default maximum slow charge cost for an EV
DEFAULT_MAX_SC_COST = 100.0

# Default minimum V2G earning for an EV
DEFAULT_MIN_V2G_EARN = 0.0


from abc import abstractmethod
from dataclasses import dataclass
import enum
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Any, Callable, Generic, Iterable, Optional, Protocol, TypeVar, runtime_checkable, List, Tuple, Dict
from feasytools import RangeList
from fpowerkit import Grid
from ..traffic import TrafficInst
from ..locale import Lang

Getter = Callable[[Any], str]
Setter = Callable[[str], Any]
Validator = Callable[[str], bool]

class PluginStatus(enum.IntEnum):
    '''Plugin status'''
    EXECUTE = 0     # Current call should execute the plugin
    HOLD = 1        # Current call should retain the result of the last plugin execution
    OFFLINE = 2     # Current call should return the return value when the plugin is offline

PIResult = TypeVar('PIResult',covariant=True)
PIExec = Callable[[int,PluginStatus],Tuple[bool,PIResult]]
PINoRet = Callable[[],None]

@dataclass
class ConfigItem:
    '''Configuration item'''
    name:str
    editor:'EditMode'
    desc:str
    default_value:Optional[Any] = None
    combo_values:Optional[List[str]] = None
    spin_range:Optional[Tuple[int,int]] = None
    rangelist_hint:bool = False
    prop_config:'Optional[ConfigItemDict]' = None

class EditMode(enum.Enum):
    DISABLED = "disabled"
    ENTRY = "entry"
    SPIN = "spin"
    COMBO = "combo"
    CHECKBOX = "checkbox"
    RANGELIST = "rangelist"
    SEGFUNC = "segfunc"
    PROP = "prop"
    PDFUNC = "pdfunc"

    @staticmethod
    def entry(): return ConfigItem("", EditMode.ENTRY, "", "")
    @staticmethod
    def spin(l:int, r:int): return ConfigItem("", EditMode.SPIN, "", spin_range=(l, r))
    @staticmethod
    def combo(values:List[Any]): return ConfigItem("", EditMode.COMBO, "", combo_values=values)
    @staticmethod
    def checkbox(checked:bool = False): return ConfigItem("", EditMode.CHECKBOX, "", default_value=checked)
    @staticmethod
    def rangelist(hint:bool = False): return ConfigItem("", EditMode.RANGELIST, "", rangelist_hint=hint)
    @staticmethod
    def segfunc(): return ConfigItem("", EditMode.SEGFUNC, "")
    @staticmethod
    def prop(prop_cfg:'ConfigItemDict'): return ConfigItem("", EditMode.PROP, "", prop_config=prop_cfg)
    @staticmethod
    def pdfunc(): return ConfigItem("", EditMode.PDFUNC, "")


class ConfigItemDict(Dict[str, ConfigItem]):
    '''Plugin configuration item dictionary'''
    def __init__(self, items: Optional[Iterable[ConfigItem]] = None, default_config_item:Optional[ConfigItem]=None):
        if items is None:
            super().__init__()
        else:
            super().__init__((x.name, x) for x in items)
        self.default_config_item = default_config_item if default_config_item is not None else ConfigItem(
            name="",
            editor=EditMode.ENTRY,
            desc="(No description)",
            default_value="",
        )

    def default_value_dict(self):
        return {k:v.default_value for k, v in self.items()}
    
    def desc_dict(self):
        return {k:v.desc for k, v in self.items()}
    
    def editor_dict(self):
        return {k:v.editor for k, v in self.items()}
    
    def get_editor(self, name:str) -> EditMode:
        return self.get(name).editor
    
    def get_desc(self, name:str) -> str:
        return self.get(name).desc
    
    def get_default_value(self, name:str) -> Any:
        return self.get(name).default_value
    
    def get(self, key: str) -> ConfigItem:
        return super().get(key, self.default_config_item)

PluginConfigItem = ConfigItem
ConfigDict = ConfigItemDict

@runtime_checkable
class IGridPlugin(Protocol):
    @property
    def Grid(self) -> Grid:
        '''Get the grid instance'''
        raise NotImplementedError

class PluginBase(Generic[PIResult]):
    __PreSimulation: Optional[PINoRet]
    __PreStep: Optional[PIExec]
    __PostSimulation: Optional[PINoRet]
    __PostStep: Optional[PIExec]
    def SetPreSimulation(self,func:PINoRet) -> None:
        '''Pre-simulation plugin processing, run after other parameters are loaded'''
        self.__PreSimulation = func
    def SetPreStep(self,func:PIExec) -> None:
        '''Plugin work before simulation step'''
        self.__PreStep = func
    def SetPostStep(self,func:PIExec) -> None:
        '''Plugin work after simulation step'''
        self.__PostStep = func
    def SetPostSimulation(self,func:PINoRet) -> None:
        '''Post-simulation plugin processing'''
        self.__PostSimulation = func

    @abstractmethod
    def _save_state(self) -> object:
        '''Save the plugin state'''
        
    @abstractmethod
    def _load_state(self,state:object) -> None:
        '''Load the plugin state'''

    @staticmethod
    @abstractmethod
    def ElemShouldHave() -> ConfigDict:
        '''Get the plugin configuration item list'''
        return ConfigDict()

    def __init__(self, inst:TrafficInst, elem:ET.Element, work_dir:Path, res_dir:Path, enable_time:Optional[RangeList]=None,
            interval:int=0, plg_deps:'Optional[List[PluginBase]]' = None, initial_state:Optional[object]=None):
        '''
        Initialize the plugin
            inst: Traffic network simulation instance
            elem: Plugin configuration XML element
            enable_time: Enable time of the plugin, if not specified, check the online subnode in xml, 
                if the online subnode does not exist, it means always enable
            interval: Plugin running interval, unit = second, 
                if not specified, the invterval attribute must be specified in xml
            plugin_dependency: Plugin dependency list
            initial_state: Initial plugin state, if specified, load the plugin state from it
        '''
        self.__PreStep = None
        self.__PostStep = None
        self.__PreSimulation = None
        self.__PostSimulation = None
        self.__lastTpre = self.__lastTpost = -1
        self.__lastOkpre = False
        self.__lastOkpost = False
        self.__name = elem.tag
        self.__interval = interval if interval > 0 else int(elem.attrib.pop("interval",0))
        if self.__interval <= 0: 
            raise ValueError(Lang.ERROR_PLUGIN_INTERVAL)
        self.__on = enable_time
        if self.__on is None:
            online_elem = elem.find("online")
            if online_elem is None: self.__on = None
            else: self.__on = RangeList(online_elem)
        if plg_deps is None: plg_deps = []

        self.__respre = self.__respost = self.Init(elem, inst, work_dir, res_dir, plg_deps)

        if initial_state is not None:
            self._load_state(initial_state)
    
    @property
    @abstractmethod
    def Description(self)->str:
        '''Get the plugin description'''
        raise NotImplementedError
    
    @property
    def Name(self)->str:
        '''Get the plugin name'''
        return self.__name
    
    @property
    def Interval(self)->int:
        '''Get the plugin running interval, unit = second'''
        return self.__interval
    
    @property
    def OnlineTime(self)->Optional[RangeList]:
        '''Get the plugin enable time'''
        return self.__on
    
    @property
    def LastTime(self)->int:
        '''
        Get the time when the last plugin was in PluginStatus.EXECUTE state
        '''
        return self.__lastTpre
    
    @property
    def LastPreStepSucceed(self)->bool:
        '''
        Get whether PreStep was successful when the last plugin was in PluginStatus.EXECUTE state
        '''
        return self.__lastOkpre
    
    @property
    def LastPostStepSucceed(self)->bool:
        '''
        Get whether PostStep was successful when the last plugin was in PluginStatus.EXECUTE state
        '''
        return self.__lastOkpost

    @property
    def LastPreStepResult(self)->PIResult:
        '''
        Get the result of PreStep when the last plugin was in PluginStatus.EXECUTE state
        '''
        return self.__respre
    
    @property
    def LastPostStepResult(self)->PIResult:
        '''
        Get the result of PostStep when the last plugin was in PluginStatus.EXECUTE state
        '''
        return self.__respost
    
    @abstractmethod
    def Init(self, elem:ET.Element, inst:TrafficInst, work_dir:Path, 
             res_dir:Path, plg_deps:'List[PluginBase]') -> PIResult:
        '''
        Initialize the plugin from the XML element, TrafficInst, work path, result path, and plugin dependency.
        Return the result of offline.
        '''
    
    def IsOnline(self, t:int):
        '''Determine if the plugin is online'''
        return self.__on is None or t in self.__on
    
    def _presim(self)->None:
        '''Run the plugin PreSimulation'''
        if self.__PreSimulation is not None:
            self.__PreSimulation()
    
    def _postsim(self)->None:
        '''Run the plugin PostSimulation'''
        if self.__PostSimulation is not None:
            self.__PostSimulation()
    
    def _precall(self, _t:int)->None:
        '''Run the plugin PreStep'''
        if self.__PreStep is None: return
        if self.__on != None and _t not in self.__on:
            self.__PreStep(_t,PluginStatus.OFFLINE)
        elif self.__lastTpre + self.__interval <= _t or self.__lastTpre < 0:
            self.__lastOkpre, self.__respre = self.__PreStep(_t,PluginStatus.EXECUTE)
            self.__lastTpre = _t
        else:
            self.__PreStep(_t,PluginStatus.HOLD)
    
    def _postcall(self, _t:int, /)->None:
        '''Run the plugin PostStep'''
        if self.__PostStep is None: return
        if self.__on != None and _t not in self.__on:
            self.__PostStep(_t,PluginStatus.OFFLINE)
        elif self.__lastTpost + self.__interval <= _t or self.__lastTpost < 0:
            self.__lastOkpost, self.__respost = self.__PostStep(_t,PluginStatus.EXECUTE)
            self.__lastTpost = _t
        else:
            self.__PostStep(_t,PluginStatus.HOLD)


from v2simux_gui.com_no_vx import *
from queue import Queue


class ProgBox(Tk): 
    def __init__(self, keys:List[str], title:str="Progress Box", width:int=300, height:int=500):
        super().__init__()
        self.title(title)
        self.geometry(f"{width}x{height}")
        
        columns = ("item", "value")
        tree = Treeview(self, columns=columns, show="headings", selectmode="none")
        tree.heading("item", text="Item")
        tree.heading("value", text="Value")
        tree.column("item", width=100)
        tree.column("value", width=200)
        tree.pack(fill=BOTH, expand=True)

        self._dict:Dict[str, Any] = {k:"" for k in keys}
        self._vals:Dict[str, Any] = {k:"" for k in keys}
        for k in keys:
            self._dict[k] = tree.insert("", "end", values=(k, ""))

        self.tree = tree
        self._Q = Queue()
        self.after(100, self._upd)

    def _upd(self):
        d:Dict[str, Tuple[int, Widget]]={}
        while not self._Q.empty():
            d.update(self._Q.get())
        for key,val in d.items():
            if not key in self._dict:
                self._dict[key] = self.tree.insert("", "end", values=(key, str(val)))
            else:
                item = self._dict[key]
                self.tree.item(item, values=(key, str(val)))
            self._vals[key] = val
        self.after(100,self._upd)
    
    def close(self):
        self.destroy()    
    
    def set_val(self, d:Dict[str, Any]):
        self._Q.put(d)


from v2simux_gui.common import *


_L = LangLib.Load(__file__)


class LoadingBox(Toplevel):
    def __init__(self, items:List[str], parentQ:EventQueue, **kwargs):
        super().__init__(None, **kwargs)
        self._pQ = parentQ
        self.title("Loading...")
        self.geometry("400x300")
        self.attributes("-topmost", True)
        self.cks:List[Label]=[]
        self.dkt:Dict[str,int]={}
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        for i, t in enumerate(items):
            Label(self, text=t).grid(column=0,row=i)
            self.cks.append(Label(self, text="..."))
            self.cks[-1].grid(column=1,row=i)
            self.dkt[t]=i
            self.rowconfigure(i, weight=1)
        self._closed = False
    
    def setText(self, itm:str, val:str):
        if self._closed: return
        self.cks[self.dkt[itm]].configure(text=val)
        for x in self.cks:
            if x['text'] != _L['DONE']: break
        else:
            self._closed = True
            self._pQ.delegate(self.destroy)


import threading
import time
import traceback
from collections import deque
from queue import Queue
from typing import Callable, Dict, Tuple

class EventQueue:
    def __init__(self, parent, maxsize:int = 0, interval_ms:int = 100, max_proc_ms:int = 90):
        self._Q = Queue(maxsize)
        self._parent = parent
        self._evt:Dict[str, Callable] = {}
        self._interval_ms = interval_ms
        self._max_proc_ns = max_proc_ms * 1e6  # Convert to nanoseconds
        self._delegates:deque[Tuple[Callable, Tuple, Dict]] = deque()
        self.do()  # Start processing events immediately
    
    def do(self):
        """Process all events in the queue."""
        prod_next = True
        st = time.time_ns()

        cnt = 0
        while self._delegates:
            func, args, kwargs = self._delegates.popleft()
            try:
                func(*args, **kwargs)
            except Exception as e:
                print(f"Error in delegate function '{func.__name__}': {e}")
        
        cnt = 0
        while not self._Q.empty() and (time.time_ns() - st < self._max_proc_ns) and cnt < 100:
            name, args, kwargs = self._Q.get()
            if name == "__quit__":
                self._Q.task_done()
                prod_next = False
                break
            if name in self._evt and self._evt[name] is not None:
                try:
                    self._evt[name](*args, **kwargs)
                except Exception as e:
                    traceback.print_exc()
                    print(f"Error processing event '{name}': {e}")
            else:
                print(f"Event '{name}' is not registered.")
            self._Q.task_done()
            self._parent.update()  # Ensure the GUI updates after processing each event
            cnt += 1
        if prod_next:
            intv = self._interval_ms
            if cnt >= 100:
                intv = max(intv // 10, 1)
            self._parent.after(intv, self.do)

    def register(self, name:str, callback:Callable):
        """Register an event handler for a specific event name."""
        if name in self._evt:
            raise ValueError(f"Event '{name}' is already registered.")
        self._evt[name] = callback
    
    def setcallback(self, name:str, callback:Callable):
        """Set or update the callback for an event."""
        if name not in self._evt:
            raise ValueError(f"Event '{name}' is not registered.")
        self._evt[name] = callback
    
    def trigger(self, name:str, *args, **kwargs):
        """Trigger an event by its name with optional arguments."""
        if name not in self._evt:
            raise ValueError(f"Event '{name}' is not registered.")
        self._Q.put((name, args, kwargs))
    
    def submit(self, name:str, func:Callable, *args, **kwargs):
        """Run a function asychoronously and submit the results to trigger an event."""
        def _run_and_trigger(name, func, *args, **kwargs):
            try:
                result = func(*args, **kwargs)
                if result is None:
                    result = ()
                elif not isinstance(result, tuple):
                    result = (result,)
                self.trigger(name, *result)
            except Exception as e:
                traceback.print_exc()
                print(f"Error in submitting function '{func.__name__}' for event '{name}': {e}")
        threading.Thread(target=_run_and_trigger, args=(name, func, *args), kwargs=kwargs).start()
    
    def asyncrun(self, func:Callable, *args, **kwargs):
        """Run a no-return function asynchronously"""
        def _run(func, *args, **kwargs):
            try:
                func(*args, **kwargs)
            except Exception as e:
                traceback.print_exc()
                print(f"Error in delegating function '{func.__name__}': {e}")
        threading.Thread(target=_run, args=(func, *args), kwargs=kwargs).start()

    def delegate(self, func:Callable, *args, **kwargs):
        """Run a no-return function on the main thread."""
        self._delegates.append((func, args, kwargs))


from v2simux_gui.common import *

from v2simux import PluginPool, PluginBase, StaPool, RangeList, load_external_components
from .controls import *
from .utils import *


_L = LangLib.Load(__file__)


class PluginEditor(ScrollableTreeView):
    def __addgetter(self):
        # 获取第1列所有值
        plgs_exist = set(self.item(i, 'values')[0] for i in self.get_children())
        plgs = [[x] for x in self.plg_pool.GetAllPlugins() if x not in plgs_exist]
        f = SelectItemDialog(plgs, _L["SIM_SELECTPLG"], [("Name", _L["PLG_NAME"])])
        f.wait_window()
        if f.selected_item is None:
            return None
        plgname = f.selected_item[0]
        plgtype = self.plg_pool.GetPluginType(plgname)
        assert issubclass(plgtype, PluginBase)
        self.setCellEditMode(plgname, "Extra", ConfigItem("Extra", EditMode.PROP, "Extra properties", prop_config=plgtype.ElemShouldHave()))
        return [plgname, 300, SIM_YES, ALWAYS_ONLINE, plgtype.ElemShouldHave().default_value_dict()]
    
    def GetEnabledPlugins(self):
        enabled_plg = []
        for i in self.get_children():
            if self.item(i, 'values')[2] == SIM_YES:
                enabled_plg.append(self.item(i, 'values')[0])
        return enabled_plg
            
    def __init__(self, master, onEnabledSet:Callable[[Tuple[Any,...], str], None] = empty_postfunc, **kwargs):
        super().__init__(master, True, True, True, True, self.__addgetter, **kwargs)
        self.sta_pool = StaPool()
        self.plg_pool = PluginPool()
        if Path(EXT_COMP).exists():
            load_external_components(EXT_COMP, self.plg_pool, self.sta_pool)
        else:
            print(f"Warning: external components folder '{EXT_COMP}' not found.")
        self["show"] = 'headings'
        self["columns"] = ("Name", "Interval", "Enabled", "Online", "Extra")
        self.column("Name", width=120, stretch=NO)
        self.column("Interval", width=100, stretch=NO)
        self.column("Enabled", width=100, stretch=NO)
        self.column("Online", width=200, stretch=NO)
        self.column("Extra", width=200, stretch=YES)
        self.heading("Name", text=_L["SIM_PLGNAME"])
        self.heading("Interval", text=_L["SIM_EXEINTV"])
        self.heading("Enabled", text=_L["SIM_ENABLED"])
        self.heading("Online", text=_L["SIM_PLGOL"])
        self.heading("Extra", text=_L["SIM_PLGPROP"])
        self.setColEditMode("Interval", ConfigItem("Interval", EditMode.SPIN, "Time interval", spin_range=(1, 86400)))
        self.setColEditMode("Enabled", ConfigItem("Enabled", EditMode.COMBO, "Enabled or not", combo_values=[SIM_YES, SIM_NO]), post_func=onEnabledSet)
        self.setColEditMode("Online", ConfigItem("Online", EditMode.RANGELIST, "Online time ranges", rangelist_hint=True))
        self.setColEditMode("Extra", ConfigItem("Extra", EditMode.DISABLED, "Extra properties"))
        self.__onEnabledSet = onEnabledSet
    
    def add(self, plg_name:str, interval:Union[int, str], enabled:str, online:Union[RangeList, str], extra:Dict[str, Any]):
        new_line = (plg_name, interval, enabled, online, str(extra))
        self.insert("", "end", values=new_line)
        plg_type = self.plg_pool.GetPluginType(plg_name)
        assert issubclass(plg_type, PluginBase)
        self.setCellEditMode(plg_name, "Extra", ConfigItem("Extra", EditMode.PROP, "Extra properties", prop_config=plg_type.ElemShouldHave()))
        self.__onEnabledSet(new_line, plg_name)
    
    def is_enabled(self, plg_name:str):
        for i in self.get_children():
            if self.item(i, 'values')[0] == plg_name:
                return self.item(i, 'values')[2] == SIM_YES
        return False       